"""The painter module provides different painters for parts of the canvas.

Painters can be swapped in and out.

Each painter takes care of a layer in the canvas (such as grid, items
and handles).
"""

from typing import Collection, Protocol

from cairo import Context as CairoContext

from gaphas.item import Item


class Painter(Protocol):
    """Painter interface."""

    def paint(self, items: Collection[Item], cairo: CairoContext) -> None:
        """Do the paint action (called from the View)."""
        pass


class ItemPainterType(Protocol):
    def paint_item(self, item: Item, cairo: CairoContext) -> None:
        """Draw a single item."""

    def paint(self, items: Collection[Item], cairo: CairoContext) -> None:
        """Do the paint action (called from the View)."""
        pass



from gaphas.solver.constraint import BaseConstraint, Constraint, MultiConstraint
from gaphas.solver.solver import Solver
from gaphas.solver.variable import (
    NORMAL,
    REQUIRED,
    STRONG,
    VERY_STRONG,
    VERY_WEAK,
    WEAK,
    Variable,
    variable,
)



"""A Canvas owns a set of Items and acts as a container for both the items and
a constraint solver.

Connections
===========

Getting Connection Information
==============================
To get connected item to a handle::

    c = canvas.connections.get_connection(handle)
    if c is not None:
        print c.connected
        print c.port
        print c.constraint


To get all connected items (i.e. items on both sides of a line)::

    classes = (i.connected for i in canvas.get_connections(item=line))


To get connecting items (i.e. all lines connected to a class)::

    lines = (c.item for c in canvas.get_connections(connected=item))
"""

from __future__ import annotations

import logging
from typing import Iterable, Protocol

import cairo

from gaphas import matrix, tree
from gaphas.connections import Connection, Connections
from gaphas.item import Item
from gaphas.model import View


def instant_cairo_context():
    """A simple Cairo context, not attached to any window."""
    surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 0, 0)
    return cairo.Context(surface)


class Canvas:
    """Container class for items."""

    def __init__(self):
        self._tree: tree.Tree[Item] = tree.Tree()
        self._connections = Connections()

        self._registered_views = set()
        self._connections.add_handler(self._on_constraint_solved)

    @property
    def solver(self):
        return self._connections.solver

    @property
    def connections(self) -> Connections:
        return self._connections

    def add(self, item, parent=None, index=None):
        """Add an item to the canvas.

        >>> c = Canvas()
        >>> from gaphas import item
        >>> i = item.Item()
        >>> c.add(i)
        >>> len(c._tree.nodes)
        1
        >>> i._canvas is c
        True
        """
        assert item not in self._tree.nodes, f"Adding already added node {item}"

        self._tree.add(item, parent, index)
        self.request_update(item)

    def _remove(self, item):
        """Remove is done in a separate, @observed, method so the undo system
        can restore removed items in the right order."""
        self._tree.remove(item)
        self._connections.disconnect_item(item)
        self._update_views(removed_items=(item,))

    def remove(self, item):
        """Remove item from the canvas.

        >>> c = Canvas()
        >>> from gaphas import item
        >>> i = item.Item()
        >>> c.add(i)
        >>> c.remove(i)
        >>> c._tree.nodes
        []
        >>> i._canvas
        """
        for child in reversed(list(self.get_children(item))):
            self.remove(child)
        self._connections.remove_connections_to_item(item)
        self._remove(item)

    def reparent(self, item, parent, index=None):
        """Set new parent for an item."""
        self._tree.move(item, parent, index)

    def get_all_items(self) -> Iterable[Item]:
        """Get a list of all items.

        >>> c = Canvas()
        >>> c.get_all_items()
        []
        >>> from gaphas import item
        >>> i = item.Item()
        >>> c.add(i)
        >>> c.get_all_items() # doctest: +ELLIPSIS
        [<gaphas.item.Item ...>]
        """
        return iter(self._tree.nodes)

    def get_root_items(self):
        """Return the root items of the canvas.

        >>> c = Canvas()
        >>> c.get_all_items()
        []
        >>> from gaphas import item
        >>> i = item.Item()
        >>> c.add(i)
        >>> ii = item.Item()
        >>> c.add(ii, i)
        >>> c.get_root_items() # doctest: +ELLIPSIS
        [<gaphas.item.Item ...>]
        """
        return self._tree.get_children(None)

    def get_parent(self, item: Item) -> Item | None:
        """See `tree.Tree.get_parent()`.

        >>> c = Canvas()
        >>> from gaphas import item
        >>> i = item.Item()
        >>> c.add(i)
        >>> ii = item.Item()
        >>> c.add(ii, i)
        >>> c.get_parent(i)
        >>> c.get_parent(ii) # doctest: +ELLIPSIS
        <gaphas.item.Item ...>
        """
        return self._tree.get_parent(item)

    def get_children(self, item: Item | None) -> Iterable[Item]:
        """See `tree.Tree.get_children()`.

        >>> c = Canvas()
        >>> from gaphas import item
        >>> i = item.Item()
        >>> c.add(i)
        >>> ii = item.Item()
        >>> c.add(ii, i)
        >>> iii = item.Item()
        >>> c.add(iii, ii)
        >>> list(c.get_children(iii))
        []
        >>> list(c.get_children(ii)) # doctest: +ELLIPSIS
        [<gaphas.item.Item ...>]
        >>> list(c.get_children(i)) # doctest: +ELLIPSIS
        [<gaphas.item.Item ...>]
        """
        return self._tree.get_children(item)

    def sort(self, items: Iterable[Item]) -> Iterable[Item]:
        """Sort a list of items in the order in which they are traversed in the
        canvas (Depth first).

        >>> c = Canvas()
        >>> from gaphas import item
        >>> i1 = item.Line()
        >>> c.add(i1)
        >>> i2 = item.Line()
        >>> c.add(i2)
        >>> i3 = item.Line()
        >>> c.add (i3)
        >>> c.update_now((i1, i2, i3)) # ensure items are indexed
        >>> s = c.sort([i2, i3, i1])
        >>> s[0] is i1 and s[1] is i2 and s[2] is i3
        True
        """
        return self._tree.order(items)

    def get_matrix_i2c(self, item: Item) -> matrix.Matrix:
        """Get the Item to Canvas matrix for ``item``.

        item:
            The item who's item-to-canvas transformation matrix should
            be found
        calculate:
            True will allow this function to actually calculate it,
            instead of raising an `AttributeError` when no matrix is
            present yet. Note that out-of-date matrices are not
            recalculated.
        """
        m = item.matrix

        parent = self._tree.get_parent(item)
        if parent is not None:
            m = m.multiply(self.get_matrix_i2c(parent))
        return m

    def request_update(self, item: Item) -> None:
        """Set an update request for the item.

        >>> c = Canvas()
        >>> from gaphas import item
        >>> i = item.Item()
        >>> ii = item.Item()
        >>> c.add(i)
        >>> c.add(ii, i)
        >>> len(c._dirty_items)
        0
        >>> c.update_now((i, ii))
        >>> len(c._dirty_items)
        0
        """
        self._update_views(dirty_items=(item,))

    def request_matrix_update(self, item):
        """Schedule only the matrix to be updated."""
        self.request_update(item)

    def update_now(self, dirty_items):
        """Perform an update of the items that requested an update."""
        try:
            # keep it here, since we need up to date matrices for the solver
            for d in dirty_items:
                d.matrix_i2c.set(*self.get_matrix_i2c(d))

            # solve all constraints
            self._connections.solve()

        except Exception as e:
            logging.error("Error while updating canvas", exc_info=e)

    def register_view(self, view: View) -> None:
        """Register a view on this canvas.

        This method is called when setting a canvas on a view and should
        not be called directly from user code.
        """
        self._registered_views.add(view)

    def unregister_view(self, view: View) -> None:
        """Unregister a view on this canvas.

        This method is called when setting a canvas on a view and should
        not be called directly from user code.
        """
        self._registered_views.discard(view)

    def _on_constraint_solved(self, cinfo: Connection) -> None:
        dirty_items = set()
        known_items = set(self._tree.nodes)
        item = cinfo.item
        if item and item in known_items:
            dirty_items.add(item)
        connected = cinfo.connected
        if connected and connected in known_items:
            dirty_items.add(connected)
        if dirty_items:
            self._update_views(dirty_items)

    def _update_views(self, dirty_items=(), removed_items=()):
        """Send an update notification to all registered views."""
        for v in self._registered_views:
            v.request_update(dirty_items, removed_items)


class Traversable(Protocol):
    def get_parent(self, item: Item) -> Item | None: ...

    def get_children(self, item: Item | None) -> Iterable[Item]: ...


def ancestors(canvas: Traversable, item: Item) -> Iterable[Item | None]:
    parent = canvas.get_parent(item)
    while parent:
        yield parent
        parent = canvas.get_parent(parent)


def all_children(canvas: Traversable, item: Item | None) -> Iterable[Item]:
    children = canvas.get_children(item)
    for child in children:
        yield child
        yield from all_children(canvas, child)



"""Basic connectors such as Ports and Handles."""

from __future__ import annotations

from gaphas.position import Position
from gaphas.solver import NORMAL
from gaphas.types import Pos, SupportsFloatPos, TypedProperty


class Handle:
    """Handles are used to support modifications of Items.

    If the handle is connected to an item, the ``connected_to``
    property should refer to the item. A ``disconnect`` handler should
    be provided that handles all disconnect behaviour (e.g. clean up
    constraints and ``connected_to``).

    Note for those of you that use the Pickle module to persist a
    canvas: The property ``disconnect`` should contain a callable
    object (with __call__() method), so the pickle handler can also
    pickle that. Pickle is not capable of pickling ``instancemethod``
    or ``function`` objects.
    """

    def __init__(
        self,
        pos: Pos = (0, 0),
        strength: int = NORMAL,
        connectable: bool = False,
        movable: bool = True,
    ) -> None:
        """Create a new handle.

        Position is in item  coordinates.
        """
        self._pos = Position(pos[0], pos[1], strength)
        self._connectable = connectable
        self._movable = movable
        self._visible = True
        self._glued = False

    def _set_pos(self, pos: Position | SupportsFloatPos) -> None:
        """
        Shortcut for ``handle.pos.pos = pos``

        >>> h = Handle((10, 10))
        >>> h.pos = (20, 15)
        >>> h.pos
        <Position object on (20, 15)>
        """
        self._pos.pos = pos

    pos: TypedProperty[Position, Position | SupportsFloatPos]
    pos = property(lambda s: s._pos, _set_pos, doc="The Handle's position")

    @property
    def connectable(self) -> bool:
        """Can this handle actually connectect to a port?"""
        return self._connectable

    @connectable.setter
    def connectable(self, connectable: bool) -> None:
        self._connectable = connectable

    @property
    def movable(self) -> bool:
        """Can this handle be moved by a mouse pointer?"""
        return self._movable

    @movable.setter
    def movable(self, movable: bool) -> None:
        self._movable = movable

    @property
    def visible(self) -> bool:
        """Is this handle visible to the user?"""
        return self._visible

    @visible.setter
    def visible(self, visible: bool) -> None:
        self._visible = visible

    @property
    def glued(self) -> bool:
        """Is the handle being moved and about to be connected?"""
        return self._glued

    @glued.setter
    def glued(self, glued: bool) -> None:
        self._glued = glued

    def __str__(self) -> str:
        return f"<{self.__class__.__name__} object on ({self._pos.x}, {self._pos.y})>"

    __repr__ = __str__



from __future__ import annotations

from gi.repository import Gdk, Gtk


class Zoom:
    def __init__(self):
        self.matrix = None
        self.x0 = 0
        self.y0 = 0
        self.sx = 1.0
        self.sy = 1.0

    def begin(self, matrix, x0, y0):
        self.matrix = matrix
        self.x0 = x0
        self.y0 = y0
        self.sx = matrix[0]
        self.sy = matrix[3]

    def update(self, scale):
        assert self.matrix
        if self.sx * scale < 0.2:
            scale = 0.2 / self.sx
        elif self.sx * scale > 20.0:
            scale = 20.0 / self.sx

        m = self.matrix
        sx = m[0]
        sy = m[3]
        ox = (m[4] - self.x0) / sx
        oy = (m[5] - self.y0) / sy
        dsx = self.sx * scale / sx
        dsy = self.sy * scale / sy
        m.translate(-ox, -oy)
        m.scale(dsx, dsy)
        m.translate(+ox, +oy)


def zoom_tools() -> (
    tuple[Gtk.GestureZoom] | tuple[Gtk.GestureZoom, Gtk.EventControllerScroll]
):
    return zoom_tool(), scroll_zoom_tool()


def zoom_tool() -> Gtk.GestureZoom:
    """Create a zoom tool as a Gtk.Gesture.

    Note: we need to keep a reference to this gesture, or else it will be destroyed.
    """
    zoom = Zoom()
    gesture = Gtk.GestureZoom.new()
    gesture.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
    gesture.connect("begin", on_begin, zoom)
    gesture.connect("scale-changed", on_scale_changed, zoom)
    return gesture


def on_begin(
    gesture: Gtk.GestureZoom,
    sequence: None,
    zoom: Zoom,
) -> None:
    view = gesture.get_widget()
    _, x0, y0 = gesture.get_point(sequence)
    zoom.begin(view.matrix, x0, y0)


def on_scale_changed(_gesture: Gtk.GestureZoom, scale: float, zoom: Zoom) -> None:
    zoom.update(scale)


def scroll_zoom_tool() -> Gtk.EventControllerScroll:
    """Ctrl-scroll wheel zoom."""
    ctrl = Gtk.EventControllerScroll.new(Gtk.EventControllerScrollFlags.BOTH_AXES)
    ctrl.connect("scroll", on_scroll)
    return ctrl


def on_scroll(controller, _dx, dy):
    view = controller.get_widget()

    modifiers = controller.get_current_event_state()

    if not modifiers & Gdk.ModifierType.CONTROL_MASK:
        return False

    # Workaround: Gtk.EventController.get_current_event() causes SEGFAULT
    view = controller.get_widget()
    x = view.get_width() / 2
    y = view.get_height() / 2
    zoom = Zoom()
    zoom.begin(view.matrix, x, y)

    zoom_factor = 0.1
    d = 1 - dy * zoom_factor
    zoom.update(d)
    return True



from __future__ import annotations

from typing import Protocol, SupportsFloat, TypeVar

# A primitive position, tuple ``(x, y)``
# Pos = Tuple[Union[float, SupportsFloat], Union[float, SupportsFloat]]
Pos = tuple[float, float]
SupportsFloatPos = tuple[SupportsFloat, SupportsFloat]

GetT = TypeVar("GetT", covariant=True)
SetT = TypeVar("SetT", contravariant=True)


class TypedProperty(Protocol[GetT, SetT]):
    def __get__(self, obj: object, type: type | None = ...) -> GetT: ...

    def __set__(self, obj: object, value: SetT) -> None: ...

    def __delete__(self, obj: object) -> None: ...



from gi.repository import Gdk, Gtk

from typing import Callable, Union

from gaphas.cursor import cursor
from gaphas.tool.itemtool import default_find_item_and_handle_at_point
from gaphas.connector import Handle
from gaphas.item import Item
from gaphas.types import Pos
from gaphas.view import GtkView


def hover_tool(
    find_item_and_handle_at_point: Callable[
        [GtkView, Pos], Union[tuple[Item, Union[Handle, None]], tuple[None, None]]
    ] = default_find_item_and_handle_at_point,
) -> Gtk.EventController:
    """Highlight the currently hovered item."""
    ctrl = Gtk.EventControllerMotion.new()
    ctrl.connect("motion", on_motion, find_item_and_handle_at_point)
    ctrl.cursor_name = ""
    return ctrl


def on_motion(ctrl, x, y, find_item_and_handle_at_point):
    view = ctrl.get_widget()
    pos = (x, y)
    item, handle = find_item_and_handle_at_point(view, pos)
    view.selection.hovered_item = item

    if item:
        v2i = view.get_matrix_v2i(item)
        pos = v2i.transform_point(x, y)

    cursor_name = cursor(item, handle, pos)
    if cursor_name != ctrl.cursor_name:
        set_cursor(view, cursor_name)
        ctrl.cursor_name = cursor_name


def set_cursor(view, cursor_name):
    cursor = Gdk.Cursor.new_from_name(cursor_name)
    view.set_cursor(cursor)



"""This module contains a connections manager."""

from __future__ import annotations

from typing import Callable, Iterator, NamedTuple

from gaphas import table
from gaphas.constraint import Constraint
from gaphas.handle import Handle
from gaphas.item import Item
from gaphas.port import Port
from gaphas.solver import Solver


class Connection(NamedTuple):
    """Information about two connected items.

    - item: connecting item
    - handle: handle of connecting item (points connected item)
    - connected: connected item
    - port: port of connected item
    - constraint: optional connection constraint
    - callback: optional disconnection callback
    """

    item: Item
    handle: Handle
    connected: Item
    port: Port
    constraint: Constraint
    callback: Callable[[Item, Handle, Item, Port], None]


class ConnectionError(Exception):
    """Exception raised when there is an error when connecting an items with
    each other."""


class Connections:
    """Manage connections and constraints."""

    def __init__(self, solver: Solver | None = None) -> None:
        self._solver = solver or Solver()
        self._connections: table.Table[Connection] = table.Table(
            Connection, tuple(range(5))
        )
        self._handlers: set[Callable[[Connection], None]] = set()

        self._solver.add_handler(self._on_constraint_solved)

    def add_handler(self, handler):
        """Add a callback handler.

        Handlers are triggered when a constraint has been solved.
        """
        self._handlers.add(handler)

    def remove_handler(self, handler):
        """Remove a previously assigned handler."""
        self._handlers.discard(handler)

    def _on_constraint_solved(self, constraint):
        for cinfo in self._connections.query(constraint=constraint):
            for handler in self._handlers:
                handler(cinfo)

    @property
    def solver(self) -> Solver:
        """The solver used by this connections instance."""
        return self._solver

    def solve(self) -> None:
        """Solve all constraints."""
        self._solver.solve()

    def add_constraint(self, item: Item, constraint: Constraint) -> Constraint:
        """Add a "simple" constraint for an item."""
        self._solver.add_constraint(constraint)
        self._connections.insert(item, None, None, None, constraint, None)
        return constraint

    def remove_constraint(self, item: Item, constraint: Constraint) -> None:
        """Remove an item specific constraint."""
        self._solver.remove_constraint(constraint)
        self._connections.delete(item, None, None, None, constraint, None)

    def connect_item(
        self,
        item: Item,
        handle: Handle,
        connected: Item,
        port: Port | None,
        constraint: Constraint | None = None,
        callback: Callable[[Item, Handle, Item, Port], None] | None = None,
    ) -> None:
        """Create a connection between two items. The connection is registered
        and the constraint is added to the constraint solver.

        The pair ``(item, handle)`` should be unique and not yet connected.

        The callback is invoked when the connection is broken.

        Args:
          item (Item): Connecting item (i.e. a line).
          handle (Handle): Handle of connecting item.
          connected (Item): Connected item (i.e. a box).
          port (Port): Port of connected item.
          constraint (Constraint): Constraint to keep the connection in place.
          callback (() -> None): Function to be called on disconnection.

        ``ConnectionError`` is raised in case handle is already registered
        on a connection.
        """
        if self.get_connection(handle):
            raise ConnectionError(
                f"Handle {handle} of item {item} is already connected"
            )

        self._connections.insert(item, handle, connected, port, constraint, callback)

        if constraint:
            self._solver.add_constraint(constraint)

    def disconnect_item(self, item: Item, handle: Handle | None = None) -> None:
        """Disconnect the connections of an item.

        If handle is not None, only the connection for that handle is
        disconnected.
        """
        for cinfo in list(self._connections.query(item=item, handle=handle)):
            self._disconnect_item(*cinfo)

    def _disconnect_item(
        self,
        item: Item,
        handle: Handle,
        connected: Item,
        port: Port,
        constraint: Constraint,
        callback: Callable[[Item, Handle, Item, Port], None],
    ) -> None:
        """Perform the real disconnect."""
        # Same arguments as connect_item, makes reverser easy
        if constraint:
            self._solver.remove_constraint(constraint)

        if callback is not None:
            callback(item, handle, connected, port)

        self._connections.delete(item, handle, connected, port, constraint, callback)

    def remove_connections_to_item(self, item: Item) -> None:
        """Remove all connections (handles connected to and constraints) for a
        specific item (to and from the item).

        This is some brute force cleanup (e.g. if constraints are
        referenced by items, those references are not cleaned up).
        """
        disconnect_item = self._disconnect_item
        # remove connections from this item
        for cinfo in list(self._connections.query(item=item)):
            disconnect_item(*cinfo)
        # remove constraints to this item
        for cinfo in list(self._connections.query(connected=item)):
            disconnect_item(*cinfo)

    def reconnect_item(
        self,
        item: Item,
        handle: Handle,
        port: Port | None = None,
        constraint: Constraint | None = None,
    ) -> None:
        """Update an existing connection.

        This is used to provide a new constraint to the connection.
        ``item`` and ``handle`` are the keys to the to-be-updated
        connection.
        """
        cinfo = self.get_connection(handle)
        if not cinfo:
            raise ValueError(
                f'No data available for item "{item}" and handle "{handle}"'
            )

        if cinfo.constraint:
            self._solver.remove_constraint(cinfo.constraint)
        self._connections.delete(item=cinfo.item, handle=cinfo.handle)

        self._connections.insert(
            item,
            handle,
            cinfo.connected,
            port or cinfo.port,
            constraint,
            cinfo.callback,
        )
        if constraint:
            self._solver.add_constraint(constraint)

    def get_connection(self, handle: Handle) -> Connection | None:
        """Get connection information for specified handle.

        >>> c = Connections()
        >>> from gaphas.item import Line
        >>> line = Line()
        >>> from gaphas import item
        >>> i = item.Line()
        >>> ii = item.Line()
        >>> c.connect_item(i, i.handles()[0], ii, ii.ports()[0])
        >>> c.get_connection(i.handles()[0])     # doctest: +ELLIPSIS
        Connection(item=<gaphas.item.Line object at 0x...)
        >>> c.get_connection(i.handles()[1])     # doctest: +ELLIPSIS
        >>> c.get_connection(ii.handles()[0])    # doctest: +ELLIPSIS
        """
        try:
            return next(self._connections.query(handle=handle))
        except StopIteration:
            return None

    def get_connections(
        self,
        item: Item | None = None,
        handle: Handle | None = None,
        connected: Item | None = None,
        port: Port | None = None,
    ) -> Iterator[Connection]:
        """Return an iterator of connection information.

        The list contains (item, handle). As a result an item may be
        in the list more than once (depending on the number of handles
        that are connected). If ``item`` is connected to itself it
        will also appear in the list.

        >>> c = Connections()
        >>> from gaphas import item
        >>> i = item.Line()
        >>> ii = item.Line()
        >>> iii = item.Line()
        >>> c.connect_item(i, i.handles()[0], ii, ii.ports()[0], None)

        >>> list(c.get_connections(item=i)) # doctest: +ELLIPSIS
        [Connection(item=<gaphas.item.Line object at 0x...]
        >>> list(c.get_connections(connected=i))
        []
        >>> list(c.get_connections(connected=ii)) # doctest: +ELLIPSIS
        [Connection(item=<gaphas.item.Line object at 0x...]

        >>> c.connect_item(ii, ii.handles()[0], iii, iii.ports()[0], None)
        >>> list(c.get_connections(item=ii)) # doctest: +ELLIPSIS
        [Connection(item=<gaphas.item.Line object at 0x...]
        >>> list(c.get_connections(connected=iii)) # doctest: +ELLIPSIS
        [Connection(item=<gaphas.item.Line object at 0x...]
        """
        return self._connections.query(
            item=item, handle=handle, connected=connected, port=port
        )



"""Allow for easily adding segments to lines."""

from __future__ import annotations

from functools import singledispatch
from typing import Union

from gaphas.connector import Handle, LinePort
from gaphas.cursor import DEFAULT_CURSOR, LINE_CURSOR, cursor, line_hover
from gaphas.geometry import distance_point_point_fast
from gaphas.item import Line, matrix_i2i
from gaphas.model import Model
from gaphas.painter.handlepainter import GREEN_4, draw_handle
from gaphas.selection import Selection
from gaphas.solver import WEAK
from gaphas.types import Pos


@singledispatch
class Segment:
    def __init__(self, item, model):
        raise TypeError

    def split_segment(self, segment, count=2): ...

    def split(self, pos): ...

    def merge_segment(self, segment, count=2): ...


@Segment.register(Line)  # type: ignore
class LineSegment:
    def __init__(self, item: Line, model: Model):
        self.item = item
        self.model = model

    def split(self, pos):
        item = self.item
        handles = item.handles()
        x, y = item.matrix_i2c.inverse().transform_point(*pos)
        for h1, h2 in zip(handles, handles[1:]):
            xp = (h1.pos.x + h2.pos.x) / 2
            yp = (h1.pos.y + h2.pos.y) / 2
            if distance_point_point_fast((x, y), (xp, yp)) <= 4:
                segment = handles.index(h1)
                handles, ports = self.split_segment(segment)
                return handles and handles[0]

    def split_segment(self, segment, count=2):
        """Split one item segment into ``count`` equal pieces.

        def split_segment(self, segment, count=2):
        Two lists are returned

          - list of created handles
          - list of created ports

          :Parameters:
           segment
              Segment number to split (starting from zero).
           count
              Amount of new segments to be created (minimum 2).
        """
        item = self.item
        if segment < 0 or segment >= len(item.ports()):
            raise ValueError("Incorrect segment")
        if count < 2:
            raise ValueError("Incorrect count of segments")

        def do_split(segment, count):
            handles = item.handles()
            p0 = handles[segment].pos
            p1 = handles[segment + 1].pos
            dx, dy = p1.x - p0.x, p1.y - p0.y
            new_h = Handle((p0.x + dx / count, p0.y + dy / count), strength=WEAK)
            item.insert_handle(segment + 1, new_h)

            p0 = LinePort(p0, new_h.pos)
            p1 = LinePort(new_h.pos, p1)
            item.remove_port(item.ports()[segment])
            item.insert_port(segment, p0)
            item.insert_port(segment + 1, p1)

            if count > 2:
                do_split(segment + 1, count - 1)

        do_split(segment, count)

        # force orthogonal constraints to be recreated
        item.update_orthogonal_constraints()

        self.recreate_constraints()

        self.model.request_update(item)
        handles = item.handles()[segment + 1 : segment + count]
        ports = item.ports()[segment : segment + count - 1]
        return handles, ports

    def merge_segment(self, segment, count=2):
        """Merge two (or more) item segments.

        Tuple of two lists is returned, list of deleted handles and
        list of deleted ports.

        :Parameters:
         segment
            Segment number to start merging from (starting from zero).
         count
            Amount of segments to be merged (minimum 2).
        """
        item = self.item
        if len(item.ports()) < 2:
            raise ValueError("Cannot merge line with one segment")
        if item.orthogonal and len(item.ports()) < 1 + count:
            raise ValueError("Cannot merge orthogonal line to one segment")
        if segment < 0 or segment >= len(item.ports()):
            raise ValueError("Incorrect segment")
        if count < 2 or segment + count > len(item.ports()):
            raise ValueError("Incorrect count of segments")

        # remove handle and ports which share position with handle
        deleted_handles = item.handles()[segment + 1 : segment + count]
        deleted_ports = item.ports()[segment : segment + count]
        for h in deleted_handles:
            item.remove_handle(h)
        for p in deleted_ports:
            item.remove_port(p)

        # create new port, which replaces old ports destroyed due to
        # deleted handle
        p1 = item.handles()[segment].pos
        p2 = item.handles()[segment + 1].pos
        port = LinePort(p1, p2)
        item.insert_port(segment, port)

        # force orthogonal constraints to be recreated
        item.update_orthogonal_constraints()

        self.recreate_constraints()
        self.model.request_update(item)

        return deleted_handles, deleted_ports

    def recreate_constraints(self):
        """Create connection constraints between connecting lines and an item.

        :Parameters:
         connected
            Connected item.
        """
        connected = self.item
        model = self.model

        def find_port(line, handle, item):
            """Glue to the closest item on the canvas.

            If the item can connect, it returns a port.
            """
            pos = matrix_i2i(line, item).transform_point(*handle.pos)
            port = None
            max_dist = 10e6
            for p in item.ports():
                pg, d = p.glue(pos)
                if d >= max_dist:
                    continue
                port = p
                max_dist = d

            return port

        for cinfo in list(model.connections.get_connections(connected=connected)):
            item, handle = cinfo.item, cinfo.handle
            port = find_port(item, handle, connected)

            constraint = port.constraint(item, handle, connected)

            model.connections.reconnect_item(item, handle, port, constraint=constraint)


def segment_tool(view):
    print(
        "WARNING: You're using segment_tool(). Remove all uses of segment_tool() from your code. It's handled by item_tool now."
    )
    return None


class LineSegmentPainter:
    """This painter draws pseudo-handles on gaphas.item.Line objects. Each line
    can be split by dragging those points, which will result in a new handle.

    ConnectHandleTool take care of performing the user interaction
    required for this feature.
    """

    def __init__(self, selection: Selection):
        self.selection = selection

    def paint(self, _items, cairo):
        selection = self.selection
        item = selection.hovered_item
        if isinstance(item, Line) and item is selection.focused_item:
            h = item.handles()
            for h1, h2 in zip(h[:-1], h[1:]):
                p1, p2 = h1.pos, h2.pos
                cx = (p1.x + p2.x) / 2
                cy = (p1.y + p2.y) / 2
                vx, vy = cairo.user_to_device(*item.matrix_i2c.transform_point(cx, cy))
                cairo.set_source_rgba(*GREEN_4, 0.4)
                draw_handle(cairo, vx, vy, size=9.0, corner=1.5)


@cursor.register
def line_segment_hover(item: Line, handle: Union[Handle, None], pos: Pos) -> str:
    if not handle:
        handles = item.handles()
        if any(
            distance_point_point_fast(
                pos, ((h1.pos.x + h2.pos.x) / 2, (h1.pos.y + h2.pos.y) / 2)
            )
            <= 4.5
            for h1, h2 in zip(handles, handles[1:])
        ):
            return LINE_CURSOR
        return DEFAULT_CURSOR
    return line_hover(item, handle, pos)



"""Basic connectors such as Ports and Handles."""

from __future__ import annotations

from typing import TYPE_CHECKING

from gaphas.constraint import Constraint, LineConstraint, PositionConstraint
from gaphas.geometry import distance_line_point, distance_point_point
from gaphas.handle import Handle
from gaphas.position import MatrixProjection, Position
from gaphas.solver import MultiConstraint
from gaphas.types import Pos, SupportsFloatPos

if TYPE_CHECKING:
    from gaphas.item import Item


class Port:
    """Port connectable part of an item.

    The Item's handle connects to a port.
    """

    def __init__(self) -> None:
        super().__init__()

        self._connectable = True

    def _set_connectable(self, connectable: bool) -> None:
        self._connectable = connectable

    connectable = property(lambda s: s._connectable, _set_connectable)

    def glue(self, pos: SupportsFloatPos) -> tuple[Pos, float]:
        """Get glue point on the port and distance to the port."""
        raise NotImplementedError("Glue method not implemented")

    def constraint(self, item: Item, handle: Handle, glue_item: Item) -> Constraint:
        """Create connection constraint between item's handle and glue item."""
        raise NotImplementedError("Constraint method not implemented")


class LinePort(Port):
    """Port defined as a line between two handles."""

    def __init__(self, start: Position, end: Position) -> None:
        super().__init__()

        self.start = start
        self.end = end

    def glue(self, pos: SupportsFloatPos) -> tuple[Pos, float]:
        """Get glue point on the port and distance to the port.

        >>> p1, p2 = (0.0, 0.0), (100.0, 100.0)
        >>> port = LinePort(p1, p2)
        >>> port.glue((50, 50))
        ((50.0, 50.0), 0.0)
        >>> port.glue((0, 10))
        ((5.0, 5.0), 7.0710678118654755)
        """
        d, pl = distance_line_point(
            self.start.tuple(), self.end.tuple(), (float(pos[0]), float(pos[1]))
        )
        return pl, d

    def constraint(self, item: Item, handle: Handle, glue_item: Item) -> Constraint:
        """Create connection line constraint between item's handle and the
        port."""
        start = MatrixProjection(self.start, glue_item.matrix_i2c)
        end = MatrixProjection(self.end, glue_item.matrix_i2c)
        point = MatrixProjection(handle.pos, item.matrix_i2c)
        line = LineConstraint((start.pos, end.pos), point.pos)
        return MultiConstraint(start, end, point, line)


class PointPort(Port):
    """Port defined as a point."""

    def __init__(self, point: Position) -> None:
        super().__init__()
        self.point = point

    def glue(self, pos: SupportsFloatPos) -> tuple[Pos, float]:
        """Get glue point on the port and distance to the port.

        >>> h = Handle((10, 10))
        >>> port = PointPort(h.pos)
        >>> port.glue((10, 0))
        (<Position object on (10, 10)>, 10.0)
        """
        point: tuple[float, float] = self.point.pos  # type: ignore[assignment]
        d = distance_point_point(point, (float(pos[0]), float(pos[1])))
        return point, d

    def constraint(
        self, item: Item, handle: Handle, glue_item: Item
    ) -> MultiConstraint:
        """Return connection position constraint between item's handle and the
        port."""
        origin = MatrixProjection(self.point, glue_item.matrix_i2c)
        point = MatrixProjection(handle.pos, item.matrix_i2c)
        c = PositionConstraint(origin.pos, point.pos)
        return MultiConstraint(origin, point, c)



from __future__ import annotations

from collections.abc import Collection

from cairo import LINE_JOIN_ROUND
from cairo import Context as CairoContext

from gaphas.item import DrawContext, Item
from gaphas.selection import Selection


class ItemPainter:
    def __init__(self, selection: Selection | None = None) -> None:
        self.selection = selection or Selection()

    def paint_item(self, item: Item, cairo: CairoContext) -> None:
        cairo.save()
        try:
            cairo.set_line_join(LINE_JOIN_ROUND)
            cairo.transform(item.matrix_i2c.to_cairo())

            selection = self.selection
            item.draw(
                DrawContext(
                    cairo=cairo,
                    selected=(item in selection.selected_items),
                    focused=(item is selection.focused_item),
                    hovered=(item is selection.hovered_item),
                )
            )

        finally:
            cairo.restore()

    def paint(self, items: Collection[Item], cairo: CairoContext) -> None:
        """Draw the items."""

        for item in items:
            self.paint_item(item, cairo)



from __future__ import annotations

from functools import singledispatch
from typing import TYPE_CHECKING, Protocol

from gaphas.item import Item
from gaphas.types import Pos

if TYPE_CHECKING:
    from gaphas.view import GtkView


class MoveType(Protocol):
    item: Item

    def __init__(self, item: Item, view: GtkView): ...

    def start_move(self, pos: Pos) -> None: ...

    def move(self, pos: Pos) -> None: ...

    def stop_move(self, pos: Pos) -> None: ...


class ItemMove:
    """Aspect for dealing with motion on an item.

    In this case the item is moved.
    """

    last_x: float
    last_y: float

    def __init__(self, item: Item, view: GtkView):
        self.item = item
        self.view = view

    def start_move(self, pos: Pos) -> None:
        self.last_x, self.last_y = pos

    def move(self, pos: Pos) -> None:
        """Move the item.

        x and y are in view coordinates.
        """
        item = self.item
        view = self.view
        assert view.model

        v2i = view.get_matrix_v2i(item)
        x, y = pos
        dx, dy = x - self.last_x, y - self.last_y
        dx, dy = v2i.transform_distance(dx, dy)
        self.last_x, self.last_y = x, y

        item.matrix.translate(dx, dy)
        view.request_update([item])

    def stop_move(self, pos: Pos) -> None:
        pass


Move = singledispatch(ItemMove)



"""Simple class containing the tree structure for the canvas items."""

from __future__ import annotations

from contextlib import suppress
from typing import Generic, Iterable, Sequence, TypeVar

T = TypeVar("T")


class Tree(Generic[T]):
    """A Tree structure. Nodes are stores in a depth-first order.

    ``None`` is the root node.

    @invariant: len(self._children) == len(self._nodes) + 1
    """

    def __init__(self) -> None:
        # List of nodes in the tree, sorted in the order they ought to be
        # rendered
        self._nodes: list[T] = []

        # Per entry a list of children is maintained.
        self._children: dict[T | None, list[T]] = {None: []}

        # For easy and fast lookups, also maintain a child -> parent mapping
        self._parents: dict[T, T] = {}

    @property
    def nodes(self) -> Sequence[T]:
        return list(self._nodes)

    def get_parent(self, node: T) -> T | None:
        """Return the parent item of ``node``.

        >>> tree = Tree()
        >>> tree.add('n1')
        >>> tree.add('n2', parent='n1')
        >>> tree.get_parent('n2')
        'n1'
        """
        return self._parents.get(node)

    def get_children(self, node: T | None) -> Iterable[T]:
        """Return all child objects of ``node``.

        >>> tree = Tree()
        >>> tree.add('n1')
        >>> tree.add('n2', parent='n1')
        >>> tree.add('n3', parent='n1')
        >>> tree.get_children('n1')
        ['n2', 'n3']
        >>> tree.get_children('n2')
        []
        """
        return self._children[node]

    def get_siblings(self, node: T) -> list[T]:
        """Get all siblings of ``node``, including ``node``.

        >>> tree = Tree()
        >>> tree.add('n1')
        >>> tree.add('n2', parent='n1')
        >>> tree.add('n3', parent='n1')
        >>> tree.get_siblings('n2')
        ['n2', 'n3']
        """
        parent = self.get_parent(node)
        return self._children[parent]

    def get_next_sibling(self, node: T) -> T:
        """Return the node on the same level after ``node``.

        >>> tree = Tree()
        >>> tree.add('n1')
        >>> tree.add('n2', parent='n1')
        >>> tree.add('n3', parent='n1')
        >>> tree.get_next_sibling('n2')
        'n3'
        >>> tree.get_next_sibling('n3') # doctest: +ELLIPSIS
        Traceback (most recent call last):
            ...
        IndexError: list index out of range
        """
        parent = self.get_parent(node)
        siblings = self._children[parent]
        return siblings[siblings.index(node) + 1]

    def get_previous_sibling(self, node: T) -> T:
        """Return the node on the same level before ``node``.

        >>> tree = Tree()
        >>> tree.add('n1')
        >>> tree.add('n2', parent='n1')
        >>> tree.add('n3', parent='n1')
        >>> tree.get_previous_sibling('n3')
        'n2'
        >>> tree.get_previous_sibling('n2') # doctest: +ELLIPSIS
        Traceback (most recent call last):
            ...
        IndexError: list index out of range
        """
        parent = self.get_parent(node)
        siblings = self._children[parent]
        index = siblings.index(node) - 1
        if index < 0:
            raise IndexError("list index out of range")
        return siblings[index]

    def get_all_children(self, node: T) -> Iterable[T]:
        """Iterate all children (and children of children and so forth)

        >>> tree = Tree()
        >>> tree.add('n1')
        >>> tree.add('n2', parent='n1')
        >>> tree.add('n3', parent='n2')
        >>> tree.get_children('n1')
        ['n2']
        >>> tree.get_all_children('n1') # doctest: +ELLIPSIS
        <generator object Tree.get_all_children at 0x...>
        >>> list(tree.get_all_children('n1'))
        ['n2', 'n3']
        """
        children = self.get_children(node)
        for c in children:
            yield c
            yield from self.get_all_children(c)

    def get_ancestors(self, node: T) -> Iterable[T]:
        """Iterate all parents and parents of parents, etc.

        >>> tree = Tree()
        >>> tree.add('n1')
        >>> tree.add('n2', parent='n1')
        >>> tree.add('n3', parent='n2')
        >>> tree.get_parent('n3')
        'n2'
        >>> tree.get_ancestors('n3') # doctest: +ELLIPSIS
        <generator object Tree.get_ancestors at 0x...>
        >>> list(tree.get_ancestors('n3'))
        ['n2', 'n1']
        >>> list(tree.get_ancestors('n1'))
        []
        """
        parent = self.get_parent(node)
        while parent:
            yield parent
            parent = self.get_parent(parent)

    def order(self, items: Iterable[T]) -> Iterable[T]:
        items_set = set(items)
        return (n for n in self._nodes if n in items_set)

    def _add_to_nodes(
        self, node: T, parent: T | None, index: int | None = None
    ) -> None:
        """Helper method to place nodes on the right location in the nodes list
        Called only from add() and move()"""
        nodes = self._nodes
        siblings = self._children[parent]
        try:
            atnode = siblings[index]  # type: ignore[index]
        except (TypeError, IndexError):
            index = len(siblings)
            if parent:
                try:
                    next_uncle = self.get_next_sibling(parent)
                except IndexError:
                    # parent has no younger brothers..
                    # place it before the next uncle of grant_parent:
                    return self._add_to_nodes(node, self.get_parent(parent))
                else:
                    nodes.insert(nodes.index(next_uncle), node)
            else:
                # append to root node:
                nodes.append(node)
        else:
            nodes.insert(nodes.index(atnode), node)

    def _add(self, node: T, parent: T | None = None, index: int | None = None) -> None:
        """Helper method for both add() and move()."""
        assert node not in self._nodes

        siblings = self._children[parent]

        self._add_to_nodes(node, parent, index)

        # Fix parent-child and child-parent relationship
        try:
            siblings.insert(index, node)  # type: ignore[arg-type]
        except TypeError:
            siblings.append(node)

        # Create new entry for it's own children:
        if parent:
            self._parents[node] = parent

    def add(self, node: T, parent: T | None = None, index: int | None = None) -> None:
        """Add node to the tree. parent is the parent node, which may be None
        if the item should be added to the root item.

        For usage, see the unit tests.
        """
        self._add(node, parent, index)
        self._children[node] = []

    def _remove(self, node: T) -> None:
        # Remove from parent item
        self.get_siblings(node).remove(node)
        # Remove data entries:
        del self._children[node]
        self._nodes.remove(node)
        with suppress(KeyError):
            del self._parents[node]

    def remove(self, node: T) -> None:
        """Remove ``node`` from the tree.

        For usage, see the unit tests.
        """
        # First remove children:
        for c in reversed(list(self._children[node])):
            self.remove(c)
        self._remove(node)

    def _reparent_nodes(self, node: T, parent: T | None) -> None:
        """Helper for move().

        The _children and _parent trees can be left intact as far as
        children of the reparented node are concerned. Only the position
        in the _nodes list changes.
        """
        self._nodes.remove(node)
        self._add_to_nodes(node, parent)
        for c in self._children[node]:
            self._reparent_nodes(c, node)

    def move(self, node: T, parent: T | None, index: int | None = None) -> None:
        """Set new parent for a ``node``. ``Parent`` can be ``None``,
        indicating it's added to the top.

        >>> tree = Tree()
        >>> tree.add('n1')
        >>> tree.add('n2', parent='n1')
        >>> tree.add('n3', parent='n1')
        >>> tree.nodes
        ['n1', 'n2', 'n3']
        >>> tree.move('n2', 'n3')
        >>> tree.get_parent('n2')
        'n3'
        >>> tree.get_children('n3')
        ['n2']
        >>> tree.nodes
        ['n1', 'n3', 'n2']

        If a node contains children, those are also moved:

        >>> tree.add('n4')
        >>> tree.nodes
        ['n1', 'n3', 'n2', 'n4']
        >>> tree.move('n1', 'n4')
        >>> tree.get_parent('n1')
        'n4'
        >>> list(tree.get_all_children('n4'))
        ['n1', 'n3', 'n2']
        >>> tree.nodes
        ['n4', 'n1', 'n3', 'n2']
        """
        if parent is self.get_parent(node):
            return

        # Remove all node references:
        old_parent = self.get_parent(node)
        self._children[old_parent].remove(node)
        self._nodes.remove(node)
        if old_parent:
            del self._parents[node]

        self._add(node, parent, index)

        # reorganize children in nodes list
        for c in self._children[node]:
            self._reparent_nodes(c, node)



from __future__ import annotations

from collections.abc import Callable

from gi.repository import Gtk

from gaphas.handlemove import HandleMove
from gaphas.item import Item
from gaphas.move import MoveType

FactoryType = Callable[[], Item]


def placement_tool(factory: FactoryType, handle_index: int) -> Gtk.GestureDrag:
    """Place a new item on the model."""
    gesture = Gtk.GestureDrag.new()
    placement_state = PlacementState(factory, handle_index)
    gesture.connect("drag-begin", on_drag_begin, placement_state)
    gesture.connect("drag-update", on_drag_update, placement_state)
    gesture.connect("drag-end", on_drag_end, placement_state)
    return gesture


class PlacementState:
    def __init__(self, factory: FactoryType, handle_index: int):
        self.factory = factory
        self.handle_index = handle_index
        self.moving: MoveType | None = None
        self.start_x = 0
        self.start_y = 0


def on_drag_begin(gesture, start_x, start_y, placement_state):
    view = gesture.get_widget()
    item = placement_state.factory()
    x, y = view.get_matrix_v2i(item).transform_point(start_x, start_y)
    item.matrix.translate(x, y)
    view.selection.unselect_all()
    view.selection.focused_item = item

    gesture.set_state(Gtk.EventSequenceState.CLAIMED)

    handle = item.handles()[placement_state.handle_index]
    if handle.movable:
        placement_state.start_x = start_x
        placement_state.start_y = start_y
        placement_state.moving = HandleMove(item, handle, view)
        placement_state.moving.start_move((start_x, start_y))


def on_drag_update(gesture, offset_x, offset_y, placement_state):
    if placement_state.moving:
        placement_state.moving.move(
            (placement_state.start_x + offset_x, placement_state.start_y + offset_y)
        )


def on_drag_end(gesture, offset_x, offset_y, placement_state):
    if placement_state.moving:
        placement_state.moving.stop_move(
            (placement_state.start_x + offset_x, placement_state.start_y + offset_y)
        )



from __future__ import annotations

from math import isclose
from typing import Callable

from gi.repository import Gtk

from gaphas.geometry import Rectangle


class Scrolling:
    """Contains Gtk.Adjustment and related code."""

    def __init__(
        self, scrolling_updated: Callable[[float | None, float | None], None]
    ) -> None:
        self._scrolling_updated = scrolling_updated
        self.hadjustment: Gtk.Adjustment | None = None
        self.vadjustment: Gtk.Adjustment | None = None
        self.hscroll_policy: Gtk.ScrollablePolicy | None = None
        self.vscroll_policy: Gtk.ScrollablePolicy | None = None
        self._hadjustment_handler_id = 0
        self._vadjustment_handler_id = 0

    def get_property(self, prop):
        if prop.name == "hadjustment":
            return self.hadjustment
        elif prop.name == "vadjustment":
            return self.vadjustment
        elif prop.name == "hscroll-policy":
            return self.hscroll_policy
        elif prop.name == "vscroll-policy":
            return self.vscroll_policy
        else:
            raise AttributeError(f"Unknown property {prop.name}")

    def set_property(self, prop, value):
        if prop.name == "hadjustment":
            if value is not None:
                if self.hadjustment and self._hadjustment_handler_id:
                    self.hadjustment.disconnect(self._hadjustment_handler_id)
                self.hadjustment = value
                self._hadjustment_handler_id = self.hadjustment.connect(
                    "value-changed", self.on_hadjustment_changed
                )
        elif prop.name == "vadjustment":
            if value is not None:
                if self.vadjustment and self._vadjustment_handler_id:
                    self.vadjustment.disconnect(self._vadjustment_handler_id)
                self.vadjustment = value
                self._vadjustment_handler_id = self.vadjustment.connect(
                    "value-changed", self.on_vadjustment_changed
                )
        elif prop.name == "hscroll-policy":
            self.hscroll_policy = value
        elif prop.name == "vscroll-policy":
            self.vscroll_policy = value
        else:
            raise AttributeError(f"Unknown property {prop.name}")

    def update_position(self, x: float, y: float) -> None:
        if self.hadjustment and not isclose(self.hadjustment.get_value(), x):
            self.hadjustment.handler_block(self._hadjustment_handler_id)
            self.hadjustment.set_value(-x)
            self.hadjustment.handler_unblock(self._hadjustment_handler_id)

        if self.vadjustment and not isclose(self.vadjustment.get_value(), y):
            self.vadjustment.handler_block(self._vadjustment_handler_id)
            self.vadjustment.set_value(-y)
            self.vadjustment.handler_unblock(self._vadjustment_handler_id)

    def update_adjustments(self, width: int, height: int, bounds: Rectangle) -> None:
        """Update scroll bar values (adjustments in GTK).

        The value will change when a scroll bar is moved.
        """
        # canvas limits (in view coordinates)
        c = Rectangle(*bounds)
        c.expand(min(width, height) / 2)
        u = c + Rectangle(width=width, height=height)

        if self.hadjustment:
            self.hadjustment.set_lower(u.x)
            self.hadjustment.set_upper(u.x1)
            self.hadjustment.set_step_increment(width // 10)
            self.hadjustment.set_page_increment(width)
            self.hadjustment.set_page_size(width)

        if self.vadjustment:
            self.vadjustment.set_lower(u.y)
            self.vadjustment.set_upper(u.y1)
            self.vadjustment.set_step_increment(height // 10)
            self.vadjustment.set_page_increment(height)
            self.vadjustment.set_page_size(height)

    def on_hadjustment_changed(self, adj):
        self._scrolling_updated(-adj.get_value(), None)

    def on_vadjustment_changed(self, adj):
        self._scrolling_updated(None, -adj.get_value())



from __future__ import annotations

import logging
from functools import singledispatch
from operator import itemgetter
from typing import TYPE_CHECKING, Iterable, Sequence, Iterator

from gaphas.connector import ConnectionSink, ConnectionSinkType, Connector
from gaphas.handle import Handle
from gaphas.item import Item
from gaphas.types import Pos

if TYPE_CHECKING:
    from gaphas.view import GtkView

log = logging.getLogger(__name__)


class ItemHandleMove:
    """Move a handle (role is applied to the handle)"""

    GLUE_DISTANCE = 10

    last_x: float
    last_y: float

    def __init__(self, item: Item, handle: Handle, view: GtkView):
        self.item = item
        self.handle = handle
        self.view = view

    def start_move(self, pos: Pos) -> None:
        self.last_x, self.last_y = pos
        model = self.view.model
        assert model
        if cinfo := model.connections.get_connection(self.handle):
            self.handle.glued = True
            model.connections.solver.remove_constraint(cinfo.constraint)

    def move(self, pos: Pos) -> None:
        item = self.item
        view = self.view
        assert view.model

        v2i = view.get_matrix_v2i(item)

        x, y = v2i.transform_point(*pos)

        self.handle.pos = (x, y)

        self.handle.glued = bool(self.glue(pos))

        # do not request matrix update as matrix recalculation will be
        # performed due to item normalization if required
        view.model.request_update(item)

    def stop_move(self, pos: Pos) -> None:
        self.handle.glued = False
        self.connect(pos)

    def glue(
        self, pos: Pos, distance: int = GLUE_DISTANCE
    ) -> ConnectionSinkType | None:
        """Glue to an item near a specific point.

        Returns a ConnectionSink or None.
        """
        item = self.item
        handle = self.handle
        view = self.view
        model = view.model
        assert model
        connections = model.connections

        if not handle.connectable:
            return None

        for connectable in item_at_point(view, pos, distance=distance, exclude=(item,)):
            connector = Connector(self.item, self.handle, connections)
            sink = ConnectionSink(connectable)

            if connector.glue(sink):
                return sink

        return None

    def connect(self, pos: Pos) -> None:
        """Connect a handle of a item to connectable item.

        Connectable item is found by `ConnectHandleTool.glue` method.

        :Parameters:
         item
            Connecting item.
         handle
            Handle of connecting item.
         pos
            Position to connect to (or near at least)
        """
        handle = self.handle
        model = self.view.model
        assert model
        connections = model.connections
        connector = Connector(self.item, handle, connections)

        if sink := self.glue(pos):
            connector.connect(sink)
        elif connections.get_connection(handle):
            connector.disconnect()

        model.request_update(self.item)


HandleMove = singledispatch(ItemHandleMove)


def item_distance(
    view: GtkView,
    pos: Pos,
    distance: float = 0.5,
    exclude: Sequence[Item] = (),
) -> Iterable[tuple[float, Item]]:
    """Return the topmost item located at ``pos`` (x, y).

    Parameters:
        - view: a view
        - pos: Position, a tuple ``(x, y)`` in view coordinates
        - selected: if False returns first non-selected item
    """
    item: Item
    vx, vy = pos
    rect = (vx - distance, vy - distance, distance * 2, distance * 2)
    for item in reversed(list(view.get_items_in_rectangle(rect))):
        if item in exclude:
            continue

        v2i = view.get_matrix_v2i(item)
        ix, iy = v2i.transform_point(vx, vy)
        d = item.point(ix, iy)
        if d is None:
            log.warning("Item distance is None for %s", item)
            continue
        if d < distance:
            yield d, item


def order_items(distance_items, key=itemgetter(0)):
    inside = []
    outside = []
    for e in distance_items:
        if key(e) > 0:
            outside.append(e)
        else:
            inside.append(e)

    inside.sort(key=key, reverse=True)
    outside.sort(key=key)
    return inside + outside


def item_at_point(
    view: GtkView,
    pos: Pos,
    distance: float = 0.5,
    exclude: Sequence[Item] = (),
) -> Iterator[Item]:
    """Return the topmost item located at ``pos`` (x, y).

    Parameters:
        - view: a view
        - pos: Position, a tuple ``(x, y)`` in view coordinates
        - selected: if False returns first non-selected item
    """
    return (
        item for _d, item in order_items(item_distance(view, pos, distance, exclude))
    )



"""This module contains several flavors of constraint classes.  Each has a
method `Constraint.solve_for(name)` and a method `Constraint.mark_dirty(v)`.
These methods are used by the constraint solver (`solver.Solver`) to set the
variables.

Variables should be of type `solver.Variable`.

See classes' documentation below for constraints description and for
examples of their usage.

EqualsConstraint
    Make 'a' and 'b' equal.
LessThanConstraint
    Ensure one variable stays smaller than another.
CenterConstraint
    Ensures a Variable is kept between two other variables.
BalanceConstraint
    Keeps three variables in line, maintaining a specific ratio.
LineConstraint
    Solves the equation where a line is connected to a line or side at
    a specific point.

New constraint class should derive from Constraint class abstract
class and implement `Constraint.solve_for(Variable)` method to update
a variable with appropriate value.
"""

from __future__ import annotations

import logging
import math

from gaphas.position import Position
from gaphas.solver import BaseConstraint, Constraint

log = logging.getLogger(__name__)


def constraint(
    *,
    horizontal: tuple[Position, Position] | None = None,
    vertical: tuple[Position, Position] | None = None,
    left_of: tuple[Position, Position] | None = None,
    above: tuple[Position, Position] | None = None,
    line: tuple[Position, tuple[Position, Position]] | None = None,
    delta: float = 0.0,
    align: float | None = None,
) -> Constraint:
    """Utility (factory) method to create item's internal constraint between
    two positions or between a position and a line.

    Position is a tuple of coordinates, i.e. ``(2, 4)``.

    Line is a tuple of positions, i.e. ``((2, 3), (4, 2))``.

    This method shall not be used to create constraints between
    two different items.

    Created constraint is returned.

    :Parameters:
        horizontal=(p1, p2)
        Keep positions ``p1`` and ``p2`` aligned horizontally.
        vertical=(p1, p2)
        Keep positions ``p1`` and ``p2`` aligned vertically.
        left_of=(p1, p2)
        Keep position ``p1`` on the left side of position ``p2``.
        above=(p1, p2)
        Keep position ``p1`` above position ``p2``.
        line=(p, l)
        Keep position ``p`` on line ``l``.
    """
    cc: Constraint | None
    if horizontal:
        p1, p2 = horizontal
        cc = EqualsConstraint(p1[1], p2[1], delta)
    elif vertical:
        p1, p2 = vertical
        cc = EqualsConstraint(p1[0], p2[0], delta)
    elif left_of:
        p1, p2 = left_of
        cc = LessThanConstraint(p1[0], p2[0], delta)
    elif above:
        p1, p2 = above
        cc = LessThanConstraint(p1[1], p2[1], delta)
    elif line:
        pos, line_l = line
        if align is None:
            cc = LineConstraint(line=line_l, point=pos)
        else:
            cc = LineAlignConstraint(line=line_l, point=pos, align=align, delta=delta)
    else:
        raise ValueError("Constraint incorrectly specified")
    return cc


# is simple abs(x - y) > EPSILON enough for canvas needs?
EPSILON = 1e-6


def _update(variable, value):
    if abs(variable.value - value) > EPSILON:
        variable.value = value


class EqualsConstraint(BaseConstraint):
    """Constraint, which ensures that two arguments ``a`` and ``b`` are equal:

        a + delta = b

    for example
    >>> from gaphas.solver import Variable
    >>> a, b = Variable(1.0), Variable(2.0)
    >>> eq = EqualsConstraint(a, b)
    >>> eq.solve_for(a)
    >>> a
    Variable(2, 20)
    >>> a.value = 10.8
    >>> eq.solve_for(b)
    >>> b
    Variable(10.8, 20)
    """

    def __init__(self, a=None, b=None, delta=0.0):
        super().__init__(a, b, delta)
        self.a = a
        self.b = b
        self.delta = delta

    def solve_for(self, var):
        assert var in (self.a, self.b, self.delta)

        _update(
            *(
                (var is self.a)
                and (self.a, self.b.value - self.delta)
                or (var is self.b)
                and (self.b, self.a.value + self.delta)
                or (self.delta, self.b.value - self.a.value)
            )
        )


class CenterConstraint(BaseConstraint):
    """Simple Constraint, takes three arguments: 'a', 'b' and center. When
    solved, the constraint ensures 'center' is located in the middle of 'a' and
    'b'.

    >>> from gaphas.solver import Variable
    >>> a, b, center = Variable(1.0), Variable(3.0), Variable()
    >>> eq = CenterConstraint(a, b, center)
    >>> eq.solve_for(a)
    >>> a
    Variable(1, 20)
    >>> center
    Variable(2, 20)
    >>> a.value = 10
    >>> eq.solve_for(b)
    >>> b
    Variable(3, 20)
    >>> center
    Variable(6.5, 20)
    """

    def __init__(self, a=None, b=None, center=None):
        super().__init__(a, b, center)
        self.a = a
        self.b = b
        self.center = center

    def solve_for(self, var):
        assert var in (self.a, self.b, self.center)

        v = (self.a.value + self.b.value) / 2.0
        _update(self.center, v)


class LessThanConstraint(BaseConstraint):
    """Ensure ``smaller`` is less than ``bigger``. The variable that is passed
    as to-be-solved is left alone (cause it is the variable that has not been
    moved lately). Instead the other variable is solved.

    >>> from gaphas.solver import Variable
    >>> a, b = Variable(3.0), Variable(2.0)
    >>> lt = LessThanConstraint(smaller=a, bigger=b)
    >>> lt.solve_for(a)
    >>> a, b
    (Variable(3, 20), Variable(3, 20))
    >>> b.value = 0.8
    >>> lt.solve_for(b)
    >>> a, b
    (Variable(0.8, 20), Variable(0.8, 20))

    Also minimal delta between two values can be set

    >>> a, b = Variable(10.0), Variable(8.0)
    >>> lt = LessThanConstraint(smaller=a, bigger=b, delta=5)
    >>> lt.solve_for(a)
    >>> a, b
    (Variable(10, 20), Variable(15, 20))
    """

    def __init__(self, smaller=None, bigger=None, delta=0.0):
        super().__init__(smaller, bigger, delta)
        self.smaller = smaller
        self.bigger = bigger
        self.delta = delta

    def solve_for(self, var):
        if self.smaller.value > self.bigger.value - self.delta:
            if var is self.smaller:
                self.bigger.value = self.smaller.value + self.delta
            elif var is self.bigger:
                self.smaller.value = self.bigger.value - self.delta
            elif var is self.delta:
                self.delta.value = self.bigger.value - self.smaller.value


class BalanceConstraint(BaseConstraint):
    """Ensure that a variable ``v`` is between values specified by ``band`` and
    in distance proportional from ``band[0]``.

    Consider

    >>> from gaphas.solver import Variable, WEAK
    >>> a, b, c = Variable(2.0), Variable(3.0), Variable(2.3, WEAK)
    >>> bc = BalanceConstraint(band=(a,b), v=c)
    >>> c.value = 2.4
    >>> c
    Variable(2.4, 10)
    >>> bc.solve_for(c)
    >>> a, b, c
    (Variable(2, 20), Variable(3, 20), Variable(2.3, 10))

    Band does not have to be ``band[0] < band[1]``

    >>> a, b, c = Variable(3.0), Variable(2.0), Variable(2.45, WEAK)
    >>> bc = BalanceConstraint(band=(a,b), v=c)
    >>> c.value = 2.50
    >>> c
    Variable(2.5, 10)
    >>> bc.solve_for(c)
    >>> a, b, c
    (Variable(3, 20), Variable(2, 20), Variable(2.45, 10))
    """

    def __init__(self, band=None, v=None, balance=None):
        super().__init__(band[0], band[1], v)
        self.band = band
        self.balance = balance
        self.v = v

        if self.balance is None:
            self.update_balance()

    def update_balance(self):
        b1, b2 = self.band
        w = b2 - b1
        self.balance = (self.v - b1) / w if w != 0 else 0

    def solve_for(self, var):
        b1, b2 = self.band
        w = b2.value - b1.value
        value = b1.value + w * self.balance
        _update(var, value)


class LineConstraint(BaseConstraint):
    """Ensure a point is kept on a line.

    Attributes:
     - _line: line defined by tuple ((x1, y1), (x2, y2))
     - _point: point defined by tuple (x, y)
    """

    def __init__(self, line, point):
        super().__init__(
            line[0][0], line[0][1], line[1][0], line[1][1], point[0], point[1]
        )

        self._line = line
        self._point = point
        self.update_ratio()

    def update_ratio(self):
        """
        >>> from gaphas.solver import Variable
        >>> line = (Variable(0), Variable(0)), (Variable(30), Variable(20))
        >>> point = (Variable(15), Variable(4))
        >>> lc = LineConstraint(line=line, point=point)
        >>> lc.update_ratio()
        >>> lc.ratio_x, lc.ratio_y
        (0.5, 0.2)
        >>> line[1][0].value = 40
        >>> line[1][1].value = 30
        >>> lc.solve_for(point[0])
        >>> lc.ratio_x, lc.ratio_y
        (0.5, 0.2)
        >>> point
        (Variable(20, 20), Variable(6, 20))
        """
        sx, sy = self._line[0]
        ex, ey = self._line[1]
        px, py = self._point

        try:
            self.ratio = float(px.value - sx.value) / float(ex.value - sx.value)
        except ZeroDivisionError:
            try:
                self.ratio = float(py.value - sy.value) / float(ey.value - sy.value)
            except ZeroDivisionError:
                self.ratio = 0.0

    def solve_for(self, var=None):
        self._solve()

    def _solve(self):
        """Solve the equation for the connected_handle.

        >>> from gaphas.solver import Variable
        >>> line = (Variable(0), Variable(0)), (Variable(30), Variable(20))
        >>> point = (Variable(15), Variable(4))
        >>> lc = LineConstraint(line=line, point=point)
        >>> lc.update_ratio()
        >>> lc.solve_for(point[0])
        >>> point
        (Variable(15, 20), Variable(4, 20))
        >>> line[1][0].value = 40
        >>> line[1][1].value =  30
        >>> lc.solve_for(point[0])
        >>> point
        (Variable(20, 20), Variable(6, 20))
        """
        sx, sy = self._line[0]
        ex, ey = self._line[1]
        px, py = self._point

        x = sx.value + (ex.value - sx.value) * self.ratio
        y = sy.value + (ey.value - sy.value) * self.ratio

        _update(px, x)
        _update(py, y)


class PositionConstraint(BaseConstraint):
    """Ensure that point is always in origin position.

    Attributes:
     - _origin: origin position
     - _point: point to be in origin position
    """

    def __init__(self, origin, point):
        super().__init__(origin[0], origin[1], point[0], point[1])

        self._origin = origin
        self._point = point

    def solve_for(self, var=None):
        """Ensure that point's coordinates are the same as coordinates of the
        origin position."""
        x, y = self._origin[0].value, self._origin[1].value
        _update(self._point[0], x)
        _update(self._point[1], y)


class LineAlignConstraint(BaseConstraint):
    """Ensure a point is kept on a line in position specified by align and
    padding information.

    Align is specified as a number between 0 and 1, for example
     0
        keep point at one end of the line
     1
        keep point at other end of the line
     0.5
        keep point in the middle of the line

    Align can be adjusted with `delta` parameter, which specifies the padding of
    the point.

    :Attributes:
     _line
        Line defined by tuple ((x1, y1), (x2, y2)).
     _point
        Point defined by tuple (x, y).
     _align
        Align of point.
     _delta
        Padding of the align.
    """

    def __init__(self, line, point, align=0.5, delta=0.0):
        super().__init__(
            line[0][0], line[0][1], line[1][0], line[1][1], point[0], point[1]
        )

        self._line = line
        self._point = point
        self._align = align
        self._delta = delta

    def solve_for(self, var=None):
        sx, sy = self._line[0]
        ex, ey = self._line[1]
        px, py = self._point
        a = math.atan2(ey.value - sy.value, ex.value - sx.value)

        x = sx.value + (ex.value - sx.value) * self._align + self._delta * math.cos(a)
        y = sy.value + (ey.value - sy.value) * self._align + self._delta * math.sin(a)

        _update(px, x)
        _update(py, y)



from __future__ import annotations

from typing import Callable, SupportsFloat

from gaphas.types import TypedProperty

# epsilon for float comparison
# is simple abs(x - y) > EPSILON enough for canvas needs?
EPSILON = 1e-6

# Variable Strengths:
VERY_WEAK = 0
WEAK = 10
NORMAL = 20
STRONG = 30
VERY_STRONG = 40
REQUIRED = 100


class variable:
    """Easy-to-use drop Variable descriptor.

    >>> class A(object):
    ...     x = variable(varname='_v_x')
    ...     y = variable(STRONG)
    ...     def __init__(self):
    ...         self.x = 12
    >>> a = A()
    >>> a.x
    Variable(12, 20)
    >>> a._v_x
    Variable(12, 20)
    >>> a.x = 3
    >>> a.x
    Variable(3, 20)
    >>> a.y
    Variable(0, 30)
    """

    def __init__(self, strength=NORMAL, varname=None):
        self._strength = strength
        self._varname = varname or f"_variable_{id(self)}"

    def __get__(self, obj, class_=None):
        if not obj:
            return self
        try:
            return getattr(obj, self._varname)
        except AttributeError:
            setattr(obj, self._varname, Variable(strength=self._strength))
            return getattr(obj, self._varname)

    def __set__(self, obj, value):
        try:
            getattr(obj, self._varname).value = float(value)
        except AttributeError:
            v = Variable(strength=self._strength)
            setattr(obj, self._varname, v)
            v.value = value


class Variable:
    """Representation of a variable in the constraint solver.

    Each Variable has a ``value`` and a ``strength``. In a constraint the weakest
    variables are changed.

    You can even do some calculating with it. The Variable always represents a
    float variable.

    The ``variable`` decorator can be used to easily define variables in classes.
    """

    def __init__(self, value: SupportsFloat = 0.0, strength: int = NORMAL):
        self._value = float(value)
        self._strength = strength
        self._handlers: set[Callable[[Variable, float], None]] = set()

    def add_handler(self, handler: Callable[[Variable, float], None]) -> None:
        """Add a handler, to be invoked when the value changes."""
        self._handlers.add(handler)

    def remove_handler(self, handler: Callable[[Variable, float], None]) -> None:
        """Remove a handler."""
        self._handlers.discard(handler)

    def notify(self, old: float) -> None:
        """Notify all handlers."""
        for handler in self._handlers:
            handler(self, old)

    @property
    def strength(self) -> int:
        """Strength."""
        return self._strength

    def dirty(self) -> None:
        """Mark the variable dirty in all attached constraints.

        Variables are marked dirty also during constraints solving to
        solve all dependent constraints, i.e. two equals constraints
        between 3 variables.
        """
        self.notify(self._value)

    def set_value(self, value: SupportsFloat) -> None:
        oldval = self._value
        v = float(value)
        if abs(oldval - v) > EPSILON:
            self._value = v
            self.notify(oldval)

    value: TypedProperty[float, SupportsFloat]
    value = property(lambda s: s._value, set_value)

    def __str__(self):
        return f"Variable({self._value:g}, {self._strength:d})"

    __repr__ = __str__

    def __float__(self):
        return float(self._value)

    def __eq__(self, other):
        """
        >>> Variable(5) == 5
        True
        >>> Variable(5) == 4
        False
        >>> Variable(5) != 5
        False
        """
        return abs(self._value - other) < EPSILON

    def __ne__(self, other):
        """
        >>> Variable(5) != 4
        True
        >>> Variable(5) != 5
        False
        """
        return abs(self._value - other) > EPSILON

    def __gt__(self, other):
        """
        >>> Variable(5) > 4
        True
        >>> Variable(5) > 5
        False
        """
        return self._value.__gt__(float(other))

    def __lt__(self, other):
        """
        >>> Variable(5) < 4
        False
        >>> Variable(5) < 6
        True
        """
        return self._value.__lt__(float(other))

    def __ge__(self, other):
        """
        >>> Variable(5) >= 5
        True
        """
        return self._value.__ge__(float(other))

    def __le__(self, other):
        """
        >>> Variable(5) <= 5
        True
        """
        return self._value.__le__(float(other))

    def __add__(self, other):
        """
        >>> Variable(5) + 4
        9.0
        """
        return self._value.__add__(float(other))

    def __sub__(self, other):
        """
        >>> Variable(5) - 4
        1.0
        >>> Variable(5) - Variable(4)
        1.0
        """
        return self._value.__sub__(float(other))

    def __mul__(self, other):
        """
        >>> Variable(5) * 4
        20.0
        >>> Variable(5) * Variable(4)
        20.0
        """
        return self._value.__mul__(float(other))

    def __floordiv__(self, other):
        """
        >>> Variable(21) // 4
        5.0
        >>> Variable(21) // Variable(4)
        5.0
        """
        return self._value.__floordiv__(float(other))

    def __mod__(self, other):
        """
        >>> Variable(5) % 4
        1.0
        >>> Variable(5) % Variable(4)
        1.0
        """
        return self._value.__mod__(float(other))

    def __divmod__(self, other):
        """
        >>> divmod(Variable(21), 4)
        (5.0, 1.0)
        >>> divmod(Variable(21), Variable(4))
        (5.0, 1.0)
        """
        return self._value.__divmod__(float(other))

    def __pow__(self, other):
        """
        >>> pow(Variable(5), 4)
        625.0
        >>> pow(Variable(5), Variable(4))
        625.0
        """
        return self._value.__pow__(float(other))

    def __truediv__(self, other):
        """
        >>> Variable(5) / 4.
        1.25
        >>> Variable(5) / Variable(4)
        1.25
        >>> Variable(5.) / 4
        1.25
        >>> 10 / Variable(5.)
        2.0
        """
        return self._value.__truediv__(float(other))

    # .. And the other way around:

    def __radd__(self, other):
        """
        >>> 4 + Variable(5)
        9.0
        >>> Variable(4) + Variable(5)
        9.0
        """
        return self._value.__radd__(float(other))

    def __rsub__(self, other):
        """
        >>> 6 - Variable(5)
        1.0
        """
        return self._value.__rsub__(other)

    def __rmul__(self, other):
        """
        >>> 4 * Variable(5)
        20.0
        """
        return self._value.__rmul__(other)

    def __rfloordiv__(self, other):
        """
        >>> 21 // Variable(4)
        5.0
        """
        return self._value.__rfloordiv__(other)

    def __rmod__(self, other):
        """
        >>> 5 % Variable(4)
        1.0
        """
        return self._value.__rmod__(other)

    def __rdivmod__(self, other):
        """
        >>> divmod(21, Variable(4))
        (5.0, 1.0)
        """
        return self._value.__rdivmod__(other)

    def __rpow__(self, other):
        """
        >>> pow(4, Variable(5))
        1024.0
        """
        return self._value.__rpow__(other)

    def __rtruediv__(self, other):
        """
        >>> 5 / Variable(4.)
        1.25
        >>> 5. / Variable(4)
        1.25
        """
        return self._value.__rtruediv__(other)



from gi.repository import Gdk, Gtk

from gaphas.tool.zoom import Zoom
from gaphas.tool.hover import set_cursor


def scroll_tools(speed: int = 10) -> Gtk.EventControllerScroll:
    return scroll_tool(speed), pan_tool()


def scroll_tool(speed: int = 10) -> Gtk.EventControllerScroll:
    """Scroll tool recognized 2 finger scroll gestures."""
    ctrl = Gtk.EventControllerScroll.new(Gtk.EventControllerScrollFlags.BOTH_AXES)
    ctrl.connect("scroll", on_scroll, speed)
    return ctrl


def on_scroll(controller, dx, dy, speed):
    view = controller.get_widget()

    modifiers = controller.get_current_event_state()

    if modifiers & Gdk.ModifierType.CONTROL_MASK:
        # Workaround: Gtk.EventController.get_current_event() causes SEGFAULT
        view = controller.get_widget()
        x = view.get_width() / 2
        y = view.get_height() / 2
        zoom = Zoom()
        zoom.begin(view.matrix, x, y)

        zoom_factor = 0.1
        d = 1 - dy * zoom_factor
        zoom.update(d)
    elif modifiers & Gdk.ModifierType.SHIFT_MASK:
        view.hadjustment.set_value(dy * speed - view.hadjustment.get_value())
        view.vadjustment.set_value(dx * speed - view.vadjustment.get_value())
    else:
        view.hadjustment.set_value(dx * speed - view.hadjustment.get_value())
        view.vadjustment.set_value(dy * speed - view.vadjustment.get_value())


class PanState:
    def __init__(self):
        self.reset()

    def reset(self):
        self.h = 0
        self.v = 0


def pan_tool() -> Gtk.GestureDrag:
    gesture = Gtk.GestureDrag.new()
    gesture.set_button(Gdk.BUTTON_MIDDLE)
    pan_state = PanState()
    gesture.connect("drag-begin", on_drag_begin, pan_state)
    gesture.connect("drag-update", on_drag_update, pan_state)
    gesture.connect("drag-end", on_drag_end)
    return gesture


def on_drag_begin(gesture, start_x, start_y, pan_state):
    view = gesture.get_widget()
    pan_state.h = view.matrix[4]
    pan_state.v = view.matrix[5]
    set_cursor(view, "move")
    gesture.set_state(Gtk.EventSequenceState.CLAIMED)


def on_drag_update(gesture, offset_x, offset_y, pan_state):
    view = gesture.get_widget()
    view.matrix.set(x0=pan_state.h + offset_x, y0=pan_state.v + offset_y)
    set_cursor(view, "move")


def on_drag_end(gesture, _offset_x, _offset_y):
    view = gesture.get_widget()
    set_cursor(view, "default")



from __future__ import annotations

from typing import Collection, Iterable, Protocol, runtime_checkable

from gaphas.connections import Connections
from gaphas.item import Item


@runtime_checkable
class View(Protocol):
    def request_update(
        self,
        items: Collection[Item],
        matrix_only_items: Collection[Item],
        removed_items: Collection[Item],
    ) -> None:
        """Propagate update requests to the view.

        By invoking this method, the View will be made aware of state changes,
        that will either:

        1. Cause the item to be fully updated
        2. Just cause the item to move, without any further updates.
        """


@runtime_checkable
class Model(Protocol):
    """Any class that adhere's to the Model protocol can be used as a model for
    GtkView."""

    @property
    def connections(self) -> Connections:
        """The connections instance used for this model."""

    def get_all_items(self) -> Iterable[Item]:
        """Iterate over all items in the order they need to be rendered in.

        Normally that will be depth-first.
        """

    def get_parent(self, item: Item) -> Item | None:
        """Get the parent item of an item.

        Returns ``None`` if there is no parent item.
        """

    def get_children(self, item: Item | None) -> Iterable[Item]:
        """Iterate all direct child items of an item."""

    def sort(self, items: Collection[Item]) -> Iterable[Item]:
        """Sort a collection of items in the order they need to be rendered
        in."""

    def request_update(self, item: Item) -> None:
        """Request update for an item.

        Arguments:
          item (Item): The item to be updated
        """

    def update_now(self, dirty_items: Collection[Item]) -> None:
        """This method is called during the update process.

        It will allow the model to do some additional updating of it's
        own.
        """

    def register_view(self, view: View) -> None:
        """Allow a view to be registered.

        Registered views should receive update requests for modified
        items.
        """

    def unregister_view(self, view: View) -> None:
        """Unregister a previously registered view.

        If a view is not registered, nothing should happen.
        """



"""This module contains everything to display a model on a screen."""

from __future__ import annotations

import asyncio
from math import isclose
from collections.abc import Collection, Iterable

import cairo
from gi.repository import Graphene, GObject, Gtk

from gaphas.geometry import Rect, Rectangle
from gaphas.item import Item
from gaphas.matrix import Matrix
from gaphas.model import Model
from gaphas.painter import DefaultPainter, ItemPainter
from gaphas.painter.painter import ItemPainterType, Painter
from gaphas.quadtree import Quadtree, QuadtreeBucket
from gaphas.selection import Selection
from gaphas.view.scrolling import Scrolling


# Handy debug flag for drawing bounding boxes around the items.
DEBUG_DRAW_BOUNDING_BOX = False
DEBUG_DRAW_QUADTREE = False

# The tolerance for Cairo. Bigger values increase speed and reduce accuracy
# (default: 0.1)
PAINT_TOLERANCE = 0.8
BOUNDING_BOX_TOLERANCE = 1.0


class GtkView(Gtk.DrawingArea, Gtk.Scrollable):
    """GTK widget for rendering a gaphas.model.Model to a screen.  The view
    uses Tools to handle events and Painters to draw. Both are configurable.

    The widget already contains adjustment objects (`hadjustment`,
    `vadjustment`) to be used for scrollbars.

    This view registers itself on the model, so it will receive
    update events.
    """

    # Just defined a name to make GTK register this class.
    __gtype_name__ = "GaphasView"

    # properties required by Gtk.Scrollable
    __gproperties__ = {
        "hscroll-policy": (
            Gtk.ScrollablePolicy,
            "hscroll-policy",
            "hscroll-policy",
            Gtk.ScrollablePolicy.MINIMUM,
            GObject.ParamFlags.READWRITE,
        ),
        "hadjustment": (
            Gtk.Adjustment,
            "hadjustment",
            "hadjustment",
            GObject.ParamFlags.READWRITE,
        ),
        "vscroll-policy": (
            Gtk.ScrollablePolicy,
            "vscroll-policy",
            "vscroll-policy",
            Gtk.ScrollablePolicy.MINIMUM,
            GObject.ParamFlags.READWRITE,
        ),
        "vadjustment": (
            Gtk.Adjustment,
            "vadjustment",
            "vadjustment",
            GObject.ParamFlags.READWRITE,
        ),
    }

    def __init__(self, model: Model | None = None, selection: Selection | None = None):
        """Create a new view.

        Args:
            model (Model): optional model to be set on construction time.
            selection (Selection): optional selection object, in case the default
                selection object (hover/select/focus) is not enough.
        """
        super().__init__()

        self._dirty_items: set[Item] = set()

        self._back_buffer: cairo.Surface | None = None
        self._back_buffer_needs_resizing = True

        self._update_task: asyncio.Task | None = None

        self._controllers: set[Gtk.EventController] = set()

        self.set_can_focus(True)
        self.set_focusable(True)
        self.connect_after("resize", GtkView.on_resize)

        def alignment_updated(x: float | None, y: float | None) -> None:
            if self._model:
                if x is not None:
                    self.matrix.set(x0=x)
                if y is not None:
                    self.matrix.set(y0=y)

        self._scrolling = Scrolling(alignment_updated)

        self._selection = selection or Selection()

        self._matrix = Matrix()
        self._painter: Painter = DefaultPainter(self)
        self._bounding_box_painter: ItemPainterType = ItemPainter(self._selection)

        # quadtree bounds are in canvas coordinates (not view!)
        self._qtree: Quadtree[Item, None] = Quadtree()

        self._model: Model | None = None
        if model:
            self.model = model

        self._selection.add_handler(self.on_selection_update)
        self._matrix.add_handler(self.on_matrix_update)

    def do_get_property(self, prop: str) -> object:
        return self._scrolling.get_property(prop)

    def do_set_property(self, prop: str, value: object) -> None:
        self._scrolling.set_property(prop, value)

    @property
    def matrix(self) -> Matrix:
        """Model root to view transformation matrix."""
        return self._matrix

    def get_matrix_i2v(self, item: Item) -> Matrix:
        """Get Item to View matrix for ``item``."""
        return item.matrix_i2c.multiply(self._matrix)

    def get_matrix_v2i(self, item: Item) -> Matrix:
        """Get View to Item matrix for ``item``."""
        m = self.get_matrix_i2v(item)
        m.invert()
        return m

    @property
    def model(self) -> Model | None:
        """The model."""
        return self._model

    @model.setter
    def model(self, model: Model | None) -> None:
        if self._model:
            self._model.unregister_view(self)
            self._selection.clear()
            self._dirty_items.clear()
            self._qtree.clear()
            if self._update_task:
                self._update_task.cancel()

        self._model = model

        if self._model:
            self._model.register_view(self)
            self.request_update(self._model.get_all_items())

    @property
    def painter(self) -> Painter:
        """Painter for drawing the view."""
        return self._painter

    @painter.setter
    def painter(self, painter: Painter) -> None:
        self._painter = painter

    @property
    def bounding_box_painter(self) -> ItemPainterType:
        """Special painter for calculating item bounding boxes."""
        return self._bounding_box_painter

    @bounding_box_painter.setter
    def bounding_box_painter(self, painter: ItemPainterType) -> None:
        self._bounding_box_painter = painter

    @property
    def selection(self) -> Selection:
        """Selected, focused and hovered items."""
        return self._selection

    @selection.setter
    def selection(self, selection: Selection) -> None:
        """Selected, focused and hovered items."""
        self._selection = selection
        if self._model:
            self.request_update(self._model.get_all_items())

    @property
    def bounding_box(self) -> Rectangle:
        """The bounding box of the complete view, relative to the view port."""
        return Rectangle(*transform_rectangle(self._matrix, self._qtree.soft_bounds))

    @property
    def hadjustment(self) -> Gtk.Adjustment:
        """Gtk adjustment object for use with a scrollbar."""
        return self._scrolling.hadjustment

    @property
    def vadjustment(self) -> Gtk.Adjustment:
        """Gtk adjustment object for use with a scrollbar."""
        return self._scrolling.vadjustment

    def clamp_item(self, item):
        """Update adjustments so the item is located inside the view port."""
        x, y, w, h = self.get_item_bounding_box(item).tuple()
        self.hadjustment.clamp_page(x, x + w)
        self.vadjustment.clamp_page(y, y + h)

    def add_controller(self, *controllers: Gtk.EventController) -> None:
        """Add a controller."""
        for controller in controllers:
            super().add_controller(controller)
        self._controllers.update(controllers)

    def remove_controller(self, controller: Gtk.EventController) -> bool:
        """Remove a controller."""
        super().remove_controller(controller)
        if controller in self._controllers:
            self._controllers.discard(controller)
            return True
        return False

    def remove_all_controllers(self) -> None:
        """Remove all registered controllers."""
        for controller in set(self._controllers):
            self.remove_controller(controller)

    def zoom(self, factor: float) -> None:
        """Zoom in/out by factor ``factor``."""
        assert self._model
        self.matrix.scale(factor, factor)
        self.request_update(self._model.get_all_items())

    def get_items_in_rectangle(
        self, rect: Rect, contain: bool = False
    ) -> Iterable[Item]:
        """Return the items in the rectangle 'rect' (in view coordinates).

        Items are automatically sorted in model's processing order.
        """
        assert self._model
        crect = transform_rectangle(self._matrix.inverse(), rect)
        items = (
            self._qtree.find_inside(crect)
            if contain
            else self._qtree.find_intersect(crect)
        )
        return self._model.sort(items)

    def get_item_bounding_box(self, item: Item) -> Rectangle:
        """Get the bounding box for the item, in view coordinates."""
        return Rectangle(
            *transform_rectangle(self._matrix, self._qtree.get_bounds(item))
        )

    def request_update(
        self,
        items: Iterable[Item],
        removed_items: Iterable[Item] = (),
    ) -> None:
        """Request update for items."""
        if items:
            self._dirty_items.update(items)

        if removed_items:
            selection = self._selection
            self._dirty_items.difference_update(removed_items)

            for item in removed_items:
                self._qtree.remove(item)
                selection.unselect_item(item)

        if items or removed_items:
            self.update()

    def update(self) -> asyncio.Task:
        """Update view status according to the items updated in the model."""

        async def _update():
            model = self._model
            if not model:
                return

            dirty_items = self.all_dirty_items()
            model.update_now(dirty_items)
            dirty_items |= self.all_dirty_items()

            old_bb = self._qtree.soft_bounds
            self.update_bounding_box(dirty_items)
            if self._qtree.soft_bounds != old_bb:
                self.update_scrolling()
            self.update_back_buffer()

        def clear_task(task):
            self._update_task = None

        if not self._update_task:
            self._update_task = asyncio.create_task(_update())
            self._update_task.add_done_callback(clear_task)
        return self._update_task

    def all_dirty_items(self) -> set[Item]:
        """Return all dirty items, clearing the marked items."""
        model = self._model
        if not model:
            return set()

        def iterate_items(items: Iterable[Item]) -> Iterable[Item]:
            assert model
            for item in items:
                parent = model.get_parent(item)
                if parent is not None and parent in items:
                    # item's matrix will be updated thanks to parent's matrix update
                    continue

                yield item

                yield from iterate_items(set(model.get_children(item)))

        dirty_items = set(iterate_items(self._dirty_items))
        self._dirty_items.clear()
        return dirty_items

    def update_bounding_box(self, items: Collection[Item]) -> None:
        """Update the bounding boxes of the model items for this view, in model
        coordinates."""
        painter = self._bounding_box_painter
        qtree = self._qtree
        for item in items:
            surface = cairo.RecordingSurface(cairo.Content.COLOR_ALPHA, None)
            cr = cairo.Context(surface)
            cr.set_tolerance(BOUNDING_BOX_TOLERANCE)

            painter.paint_item(item, cr)
            x, y, w, h = surface.ink_extents()

            qtree.add(item=item, bounds=(x, y, w, h))

    def update_scrolling(self) -> None:
        matrix = Matrix(*self._matrix)  # type: ignore[misc]
        # When zooming the bounding box to view size, disregard current scroll offsets
        matrix.set(x0=0, y0=0)
        bounds = Rectangle(*transform_rectangle(matrix, self._qtree.soft_bounds))

        self._scrolling.update_adjustments(self.get_width(), self.get_height(), bounds)

    def _debug_draw_bounding_box(self, cr, width, height):
        for item in self.get_items_in_rectangle((0, 0, width, height)):
            try:
                b = self.get_item_bounding_box(item)
            except KeyError:
                pass  # No bounding box right now..
            else:
                cr.save()
                cr.identity_matrix()
                cr.set_source_rgb(0.8, 0, 0)
                cr.set_line_width(1.0)
                cr.rectangle(*b)
                cr.stroke()
                cr.restore()

    def _debug_draw_quadtree(self, cr):
        def draw_qtree_bucket(bucket: QuadtreeBucket) -> None:
            cr.rectangle(*bucket.bounds)
            cr.stroke()
            for b in bucket._buckets:
                draw_qtree_bucket(b)

        cr.set_source_rgb(0, 0, 0.8)
        cr.set_line_width(1.0)
        draw_qtree_bucket(self._qtree._bucket)

    def do_realize(self) -> None:
        Gtk.DrawingArea.do_realize(self)

        if self._model:
            # Ensure updates are propagated
            self._model.register_view(self)
            self.request_update(self._model.get_all_items())

    def do_unrealize(self) -> None:
        if self._model:
            self._model.unregister_view(self)

        Gtk.DrawingArea.do_unrealize(self)

    def on_selection_update(self, item: Item | None) -> None:
        if self._model:
            if item is None:
                self.request_update(self._model.get_all_items())
            elif item in self._model.get_all_items():
                self.request_update((item,))

    def on_matrix_update(self, matrix, old_matrix_values):
        # Test if scale or rotation changed
        if not all(map(isclose, matrix, old_matrix_values[:4])):
            self.update_scrolling()
        self._scrolling.update_position(matrix[4], matrix[5])
        self.update_back_buffer()

    def on_resize(self, _width: int, _height: int) -> None:
        self.update_scrolling()
        if self.get_realized():
            self._back_buffer_needs_resizing = True
            self.update_back_buffer()
        else:
            self._back_buffer = None

    def update_back_buffer(self) -> None:
        self.queue_draw()

    def do_snapshot(self, snapshot):
        if self.model:
            width = self.get_width()
            height = self.get_height()
            r = Graphene.Rect()
            r.init(0, 0, width, height)
            cr = snapshot.append_cairo(r)
            cr.set_matrix(self.matrix.to_cairo())
            cr.save()
            cr.set_tolerance(PAINT_TOLERANCE)
            items = self.get_items_in_rectangle((0, 0, width, height))
            self.painter.paint(list(items), cr)
            cr.restore()

            if DEBUG_DRAW_BOUNDING_BOX:
                self._debug_draw_bounding_box(cr, width, height)

            if DEBUG_DRAW_QUADTREE:
                self._debug_draw_quadtree(cr)


def transform_rectangle(matrix: Matrix, rect: Rect) -> Rect:
    x, y, w, h = rect

    return matrix.transform_point(x, y) + matrix.transform_distance(w, h)



from __future__ import annotations

import logging
from typing import Callable, Union

from gi.repository import Gdk, Gtk

from gaphas.canvas import ancestors
from gaphas.connector import Handle
from gaphas.geometry import distance_line_point, distance_point_point_fast
from gaphas.handlemove import HandleMove, item_at_point
from gaphas.item import Item
from gaphas.move import Move
from gaphas.segment import Segment
from gaphas.types import Pos
from gaphas.view import GtkView

log = logging.getLogger(__name__)


def default_find_item_and_handle_at_point(
    view: GtkView, pos: Pos
) -> Union[tuple[Item, Union[Handle, None]], tuple[None, None]]:
    item, handle = handle_at_point(view, pos)
    return item or next(item_at_point(view, pos), None), handle  # type: ignore[return-value]


def item_tool(
    find_item_and_handle_at_point: Callable[
        [GtkView, Pos], tuple[Item, Handle | None] | tuple[None, None]
    ] = default_find_item_and_handle_at_point,
) -> Gtk.GestureDrag:
    """Handle item movement and movement of handles."""
    gesture = Gtk.GestureDrag.new()
    drag_state = DragState()
    gesture.connect(
        "drag-begin", on_drag_begin, drag_state, find_item_and_handle_at_point
    )
    gesture.connect("drag-update", on_drag_update, drag_state)
    gesture.connect("drag-end", on_drag_end, drag_state)
    return gesture


class DragState:
    def __init__(self):
        self.reset()

    def reset(self):
        self.moving_items = set()
        self.moving_handle = None
        self.start_x = 0
        self.start_y = 0

    @property
    def moving(self):
        yield from self.moving_items
        if self.moving_handle:
            yield self.moving_handle


def on_drag_begin(gesture, start_x, start_y, drag_state, find_item_and_handle_at_point):
    view = gesture.get_widget()
    pos = (start_x, start_y)
    selection = view.selection
    modifiers = gesture.get_current_event_state()
    item, handle = find_item_and_handle_at_point(view, pos)

    # Deselect all items unless CTRL or SHIFT is pressed
    # or the item is already selected.
    if not (
        modifiers & (Gdk.ModifierType.CONTROL_MASK | Gdk.ModifierType.SHIFT_MASK)
        or item in selection.selected_items
    ):
        selection.unselect_all()

    if not item:
        gesture.set_state(Gtk.EventSequenceState.DENIED)
        return

    if (
        not handle
        and item in selection.selected_items
        and modifiers & Gdk.ModifierType.CONTROL_MASK
    ):
        selection.unselect_item(item)
        gesture.set_state(Gtk.EventSequenceState.DENIED)
        return

    if not handle and item is view.selection.focused_item:
        handle = maybe_split_segment(view, item, pos)

    selection.focused_item = item
    gesture.set_state(Gtk.EventSequenceState.CLAIMED)

    drag_state.start_x = start_x
    drag_state.start_y = start_y

    if handle:
        drag_state.moving_handle = HandleMove(item, handle, view)
    else:
        drag_state.moving_items = set(moving_items(view))

    for moving in drag_state.moving:
        moving.start_move((start_x, start_y))


def moving_items(view):
    """Filter the items that should eventually be moved.

    Returns Move aspects for the items.
    """
    selected_items = set(view.selection.selected_items)
    for item in selected_items:
        # Do not move subitems of selected items
        if not set(ancestors(view.model, item)).intersection(selected_items):
            yield Move(item, view)


def on_drag_update(gesture, offset_x, offset_y, drag_state):
    view = gesture.get_widget()
    x = drag_state.start_x + offset_x
    y = drag_state.start_y + offset_y

    for moving in drag_state.moving:
        moving.move((x, y))

    if not (0 <= x <= view.get_width() and 0 <= y <= view.get_height()):
        view.clamp_item(view.selection.focused_item)


def on_drag_end(gesture, offset_x, offset_y, drag_state):
    for moving in drag_state.moving:
        moving.stop_move((drag_state.start_x + offset_x, drag_state.start_y + offset_y))
    if drag_state.moving_handle:
        moving = drag_state.moving_handle
        maybe_merge_segments(gesture.get_widget(), moving.item, moving.handle)
    drag_state.reset()


def order_handles(handles):
    if handles:
        yield handles[0]
        yield handles[-1]
        yield from handles[1:-1]


def handle_at_point(
    view: GtkView, pos: Pos, distance: int = 6
) -> tuple[Item, Handle] | tuple[None, None]:
    """Look for a handle at ``pos`` and return the tuple (item, handle)."""

    def find(item):
        """Find item's handle at pos."""
        v2i = view.get_matrix_v2i(item)
        d = distance_point_point_fast(v2i.transform_distance(0, distance))
        x, y = v2i.transform_point(*pos)

        for h in order_handles(item.handles()):
            if not h.movable:
                continue
            hx, hy = h.pos
            if -d < (hx - x) < d and -d < (hy - y) < d:
                return h

    selection = view.selection

    # The focused item is the preferred item for handle grabbing
    if selection.focused_item:
        h = find(selection.focused_item)
        if h:
            return selection.focused_item, h

    # then try hovered item
    if selection.hovered_item:
        h = find(selection.hovered_item)
        if h:
            return selection.hovered_item, h

    # Last try all items, checking the bounding box first
    x, y = pos
    items = reversed(
        list(
            view.get_items_in_rectangle(
                (x - distance, y - distance, distance * 2, distance * 2)
            )
        )
    )

    for item in items:
        h = find(item)
        if h:
            return item, h
    return None, None


def maybe_split_segment(view, item, pos):
    try:
        segment = Segment(item, view.model)
    except TypeError:
        return None
    else:
        cpos = view.matrix.inverse().transform_point(*pos)
        return segment.split(cpos)


def maybe_merge_segments(view, item, handle):
    handles = item.handles()

    # don't merge using first or last handle
    if handles[0] is handle or handles[-1] is handle:
        return

    # ensure at least three handles
    handle_index = handles.index(handle)
    segment = handle_index - 1

    # cannot merge starting from last segment
    if segment == len(item.ports()) - 1:
        segment = -1
    assert segment >= 0 and segment < len(item.ports()) - 1

    before = handles[handle_index - 1]
    after = handles[handle_index + 1]
    d, p = distance_line_point(before.pos, after.pos, handle.pos)

    if d > 4:
        return

    try:
        Segment(item, view.model).merge_segment(segment)
    except ValueError:
        pass
    else:
        if handle:
            view.model.request_update(item)



"""Module implements guides when moving items and handles around."""

from __future__ import annotations

from dataclasses import dataclass
from functools import singledispatch
from itertools import chain
from typing import TYPE_CHECKING, Iterable, SupportsFloat

from gaphas.canvas import all_children
from gaphas.connector import Handle
from gaphas.handlemove import HandleMove, ItemHandleMove
from gaphas.item import Element, Item, Line
from gaphas.move import ItemMove, Move
from gaphas.types import Pos

if TYPE_CHECKING:
    from gaphas.view import GtkView


class ItemGuide:
    """Get edges on an item, on which we can align the items."""

    def __init__(self, item: Item, handle: Handle | None = None):
        self.item = item
        self.handle = handle

    def horizontal(self) -> Iterable[SupportsFloat]:
        """Return horizontal edges (on y axis) in item coordinates."""
        return ()

    def vertical(self) -> Iterable[SupportsFloat]:
        """Return vertical edges (on x axis) in item coordinates."""
        return ()


Guide = singledispatch(ItemGuide)


@Guide.register(Element)
class ElementGuide(ItemGuide):
    """Guide to align Element items."""

    def horizontal(self) -> Iterable[SupportsFloat]:
        element = self.item
        assert isinstance(element, Element)
        if self.handle in element.handles():
            return ()
        y = element.handles()[0].pos.y
        h = element.height
        return [y, y + h / 2, y + h]

    def vertical(self) -> Iterable[SupportsFloat]:
        element = self.item
        assert isinstance(element, Element)
        if self.handle in element.handles():
            return ()
        x = element.handles()[0].pos.x
        w = element.width
        return [x, x + w / 2, x + w]


@Guide.register(Line)
class LineGuide(ItemGuide):
    """Guide for orthogonal lines."""

    def horizontal(self) -> Iterable[SupportsFloat]:
        line = self.item
        assert isinstance(line, Line)
        if not line.orthogonal or self.handle not in line.handles():
            yield from (h.pos.y for h in line.handles() if h is not self.handle)

    def vertical(self) -> Iterable[SupportsFloat]:
        line = self.item
        assert isinstance(line, Line)
        if not line.orthogonal or self.handle not in line.handles():
            yield from (h.pos.x for h in line.handles() if h is not self.handle)


@dataclass(frozen=True)
class Guides:
    vertical: list[float]
    horizontal: list[float]


MARGIN = 4


def find_vertical_guides(view, handle, item_vedges, height, excluded_items, margin):
    items = (
        set(
            chain.from_iterable(
                view.get_items_in_rectangle((x - margin, 0, margin * 2, height))
                for x in item_vedges
            )
        )
        - excluded_items
    )
    guides = (Guide(item, handle) for item in items)
    i2v = view.get_matrix_i2v
    vedges = {
        i2v(g.item).transform_point(x, 0)[0] for g in guides for x in g.vertical()
    }
    dx, edges_x = find_closest(item_vedges, vedges)
    return dx, edges_x


def find_horizontal_guides(view, handle, item_hedges, width, excluded_items, margin):
    items = (
        set(
            chain.from_iterable(
                view.get_items_in_rectangle((0, y - margin, width, margin * 2))
                for y in item_hedges
            )
        )
        - excluded_items
    )
    guides = (Guide(item, handle) for item in items)
    i2v = view.get_matrix_i2v
    hedges = {
        i2v(g.item).transform_point(0, y)[1] for g in guides for y in g.horizontal()
    }
    dy, edges_y = find_closest(item_hedges, hedges)
    return dy, edges_y


def get_excluded_items(view, item):
    """Get a set of items excluded from guide calculation."""
    assert view.model

    excluded_items = set(all_children(view.model, item))
    excluded_items.add(item)
    excluded_items.update(view.selection.selected_items)
    return excluded_items


def find_closest(item_edges, edges, margin=MARGIN):
    delta = 0
    min_d = 1000
    closest = []
    for e in edges:
        for ie in item_edges:
            d = abs(e - ie)
            if d < min_d:
                min_d = d
                delta = e - ie
                closest = [e]
            elif d == min_d:
                closest.append(e)

    return (delta, closest) if min_d <= margin else (0, ())


def update_guides(view, handle, vedges, hedges, excluded_items=frozenset()):
    dx, edges_x = find_vertical_guides(
        view, handle, vedges, view.get_height(), excluded_items, MARGIN
    )
    dy, edges_y = find_horizontal_guides(
        view, handle, hedges, view.get_width(), excluded_items, MARGIN
    )

    view.guides = Guides(edges_x, edges_y)

    view.update_back_buffer()

    return dx, dy


def reset_guides(view):
    try:
        del view.guides
    except AttributeError:
        # No problem if guides do not exist.
        pass
    else:
        view.update_back_buffer()


class GuidedItemMoveMixin:
    """Move the item, lock position on any element that's located at the same
    location."""

    view: GtkView
    item: Item
    last_x: float
    last_y: float

    def move(self, pos: Pos) -> None:
        item = self.item
        view = self.view

        px, py = pos
        pdx, pdy = px - self.last_x, py - self.last_y

        transform = view.get_matrix_i2v(item).transform_point
        item_guide = Guide(item)
        item_vedges = [transform(x, 0)[0] + pdx for x in item_guide.vertical()]
        item_hedges = [transform(0, y)[1] + pdy for y in item_guide.horizontal()]

        excluded_items = get_excluded_items(view, item)
        dx, dy = update_guides(view, None, item_vedges, item_hedges, excluded_items)

        # Call super class, with new position
        super().move((pos[0] + dx, pos[1] + dy))  # type: ignore[misc]

    def stop_move(self, pos: Pos) -> None:
        super().stop_move(pos)  # type: ignore[misc]
        reset_guides(self.view)


@Move.register(Element)
class GuidedItemMove(GuidedItemMoveMixin, ItemMove):
    pass


class GuidedItemHandleMoveMixin:
    """Move a handle and lock the position of other elements.

    Locks the position of another element that's located at the same
    position.
    """

    view: GtkView
    item: Item
    handle: Handle

    def move(self, pos: Pos) -> None:
        super().move(pos)  # type: ignore[misc]

        item = self.item
        view = self.view
        model = view.model
        assert model

        x, y = pos
        dx, dy = update_guides(view, self.handle, (x,), (y,))

        self.handle.pos = view.get_matrix_v2i(item).transform_point(x + dx, y + dy)
        model.request_update(item)

    def stop_move(self, pos: Pos) -> None:
        super().stop_move(pos)  # type: ignore[misc]
        reset_guides(self.view)


@HandleMove.register(Element)
class GuidedElementHandleMove(GuidedItemHandleMoveMixin, ItemHandleMove):
    pass


@HandleMove.register(Line)
class GuidedLineHandleMove(GuidedItemHandleMoveMixin, ItemHandleMove):
    pass


class GuidePainter:
    def __init__(self, view: GtkView):
        self.view = view

    def paint(self, _items, cr):
        try:
            guides = self.view.guides
        except AttributeError:
            return

        view = self.view

        cr.save()
        try:
            cr.identity_matrix()
            cr.set_line_width(1)
            cr.set_source_rgba(0.0, 0.0, 1.0, 0.6)
            for g in guides.vertical:
                cr.move_to(g, 0)
                cr.line_to(g, view.get_height())
                cr.stroke()
            for g in guides.horizontal:
                cr.move_to(0, g)
                cr.line_to(view.get_width(), g)
                cr.stroke()
        finally:
            cr.restore()



from __future__ import annotations

from functools import singledispatch
from typing import Callable, Protocol

from gaphas.connections import Connections
from gaphas.geometry import intersect_line_line
from gaphas.handle import Handle
from gaphas.item import Element, Item, Line, matrix_i2i
from gaphas.port import LinePort, PointPort, Port  # noqa F401
from gaphas.position import Position  # noqa F401
from gaphas.solver import Constraint
from gaphas.types import Pos, SupportsFloatPos


class ConnectionSinkType(Protocol):
    item: Item
    port: Port | None

    def __init__(self, item: Item, distance: float = 10): ...

    def glue(
        self, pos: SupportsFloatPos, secondary_pos: SupportsFloatPos | None = None
    ) -> Pos | None: ...

    def constraint(self, item: Item, handle: Handle) -> Constraint: ...


class ItemConnector:
    """Connect or disconnect an item's handle to another item or port."""

    GLUE_DISTANCE = 10  # Glue distance in view points

    def __init__(self, item: Item, handle: Handle, connections: Connections):
        self.item = item
        self.handle = handle
        self.connections = connections

    def allow(self, sink):
        return True

    def secondary_handle(self) -> Handle | None:
        return None

    def glue(self, sink: ConnectionSinkType) -> Pos | None:
        """Glue the Connector handle on the sink's port."""
        handle = self.handle
        item = self.item
        matrix = matrix_i2i(item, sink.item)
        pos = matrix.transform_point(*handle.pos)
        secondary_handle = self.secondary_handle()
        secondary_pos = (
            matrix.transform_point(*secondary_handle.pos) if secondary_handle else None
        )
        glue_pos = sink.glue(pos, secondary_pos)
        if glue_pos and self.allow(sink):
            matrix.invert()
            new_pos = matrix.transform_point(*glue_pos)
            handle.pos = new_pos
            return new_pos
        return None

    def connect(self, sink: ConnectionSinkType) -> None:
        """Connect the handle to a sink (item, port).

        Note that connect() also takes care of disconnecting in case a
        handle is reattached to another element.
        """

        # Already connected? disconnect first.
        if self.connections.get_connection(self.handle):
            self.disconnect()

        if not self.glue(sink):
            return

        self.connect_handle(sink)

    def connect_handle(
        self,
        sink: ConnectionSinkType,
        callback: Callable[[Item, Handle, Item, Port], None] | None = None,
    ) -> None:
        """Create constraint between handle of a line and port of connectable
        item.

        :Parameters:
         sink
            Connectable item and port.
         callback
            Function to be called on disconnection.
        """
        handle = self.handle
        item = self.item

        constraint = sink.constraint(item, handle)

        self.connections.connect_item(
            item, handle, sink.item, sink.port, constraint, callback=callback
        )

    def disconnect(self) -> None:
        """Disconnect the handle from the attached element."""
        self.connections.disconnect_item(self.item, self.handle)


Connector = singledispatch(ItemConnector)


@Connector.register(Line)
class LineConnector(ItemConnector):
    def secondary_handle(self) -> Handle | None:
        item = self.item
        handle = self.handle
        handles = item.handles()
        if len(handles) < 2:
            return None
        if handle is handles[0]:
            return handles[1]
        return handles[-2] if handle is handles[-1] else None


class ItemConnectionSink:
    """Makes an item a sink.

    A sink is another item that an item's handle is connected to like a
    connectable item or port.
    """

    def __init__(self, item: Item, distance: float = 10) -> None:
        self.item = item
        self.distance = distance
        self.port: Port | None = None

    def glue(
        self, pos: SupportsFloatPos, secondary_pos: SupportsFloatPos | None = None
    ) -> Pos | None:
        max_dist = self.distance
        glue_pos = None
        for p in self.item.ports():
            if not p.connectable:
                continue

            g, d = p.glue(pos)

            if d < max_dist:
                max_dist = d
                self.port = p
                glue_pos = g
        return glue_pos

    def constraint(self, item: Item, handle: Handle) -> Constraint:
        assert self.port, "constraint() can only be called after glue()"
        return self.port.constraint(item, handle, self.item)


ConnectionSink = singledispatch(ItemConnectionSink)


@ConnectionSink.register(Element)
class ElementConnectionSink(ItemConnectionSink):
    def glue(
        self, pos: SupportsFloatPos, secondary_pos: SupportsFloatPos | None = None
    ) -> Pos | None:
        if glue_pos := super().glue(pos, secondary_pos):
            return glue_pos

        if secondary_pos:
            for p in self.item.ports()[:4]:
                assert isinstance(p, LinePort)
                if point_on_line := intersect_line_line(
                    pos,  # type: ignore[arg-type]
                    secondary_pos,  # type: ignore[arg-type]
                    p.start,  # type: ignore[arg-type]
                    p.end,  # type: ignore[arg-type]
                ):
                    return point_on_line

        return None



from __future__ import annotations

from functools import singledispatch
from typing import Union

from gaphas.connector import Handle
from gaphas.item import Element, Item, Line
from gaphas.types import Pos

DEFAULT_CURSOR = "default"


@singledispatch
def cursor(item: Union[Item, None], handle: Union[Handle, None], pos: Pos) -> str:
    return DEFAULT_CURSOR


ELEMENT_CURSORS = ("nw-resize", "ne-resize", "se-resize", "sw-resize")


@cursor.register
def element_hover(item: Element, handle: Union[Handle, None], pos: Pos) -> str:
    if handle:
        index = item.handles().index(handle)
        return ELEMENT_CURSORS[index] if index < 4 else DEFAULT_CURSOR
    return DEFAULT_CURSOR


LINE_CURSOR = "move"


@cursor.register
def line_hover(item: Line, handle: Union[Handle, None], pos: Pos) -> str:
    return LINE_CURSOR if handle else DEFAULT_CURSOR



from __future__ import annotations

from typing import TYPE_CHECKING, Collection

from cairo import Context as CairoContext

from gaphas.item import Item

if TYPE_CHECKING:
    from gaphas.view import GtkView


# Colors from the GNOME Palette
RED_4 = (0.753, 0.110, 0.157)
ORANGE_4 = (0.902, 0.380, 0)
GREEN_4 = (0.180, 0.7608, 0.494)
BLUE_4 = (0.110, 0.443, 0.847)


class HandlePainter:
    """Draw handles of items that are marked as selected in the view."""

    def __init__(self, view: GtkView) -> None:
        assert view
        self.view = view

    def _draw_handles(
        self,
        item: Item,
        cairo: CairoContext,
        opacity: float | None = None,
    ) -> None:
        """Draw handles for an item.

        The handles are drawn in non-antialiased mode for clarity.
        """
        view = self.view
        model = view.model
        assert model
        cairo.save()
        if not opacity:
            opacity = 0.9 if item is view.selection.focused_item else 0.6

        get_connection = model.connections.get_connection
        for h in item.handles():
            if not h.visible:
                continue
            # connected and not being moved, see HandleTool.on_button_press
            if get_connection(h):
                color = RED_4
            elif h.glued:
                color = ORANGE_4
            elif h.movable:
                color = GREEN_4
            else:
                color = BLUE_4

            vx, vy = cairo.user_to_device(*item.matrix_i2c.transform_point(*h.pos))
            cairo.set_source_rgba(*color, opacity)

            draw_handle(cairo, vx, vy)

        cairo.restore()

    def paint(self, items: Collection[Item], cairo: CairoContext) -> None:
        view = self.view
        model = view.model
        assert model
        selection = view.selection
        # Order matters here:
        for item in model.sort(selection.selected_items):
            self._draw_handles(item, cairo)
        # Draw nice opaque handles when hovering an item:
        hovered = selection.hovered_item
        if hovered and hovered not in selection.selected_items:
            self._draw_handles(hovered, cairo, opacity=0.25)


def draw_handle(
    cairo: CairoContext, vx: float, vy: float, size: float = 12.0, corner: float = 2.0
) -> None:
    """Draw a handle with rounded corners."""
    radius = size / 2.0
    lower_right = size - corner

    pi_05 = 0.5 * 3.142
    pi = 3.142
    pi_15 = 1.5 * 3.142

    cairo.save()
    cairo.identity_matrix()
    cairo.translate(vx - radius, vy - radius)

    cairo.move_to(0.0, corner)
    cairo.arc(corner, corner, corner, pi, pi_15)
    cairo.line_to(lower_right, 0.0)
    cairo.arc(lower_right, corner, corner, pi_15, 0)
    cairo.line_to(size, lower_right)
    cairo.arc(lower_right, lower_right, corner, 0, pi_05)
    cairo.line_to(corner, size)
    cairo.arc(corner, lower_right, corner, pi_05, pi)
    cairo.close_path()
    cairo.fill()
    cairo.restore()



"""Tools provide interactive behavior to a `View` by handling specific events
sent by view.

Some of implemented tools are

`HoverTool`
    make the item under the mouse cursor the "hovered item"

`ItemTool`
    handle selection and movement of items

`HandleTool`
    handle selection and movement of handles

`RubberbandTool`
    for rubber band selection of multiple items

`PanTool`
    for easily moving the canvas around

`PlacementTool`
    for placing items on the canvas

The tools are chained with `ToolChain` class (it is a tool as well),
which allows to combine functionality provided by different tools.

Tools can handle events in different ways

- event can be ignored
- tool can handle the event (obviously)
"""
from gaphas.tool.hover import hover_tool
from gaphas.tool.itemtool import item_tool
from gaphas.tool.placement import placement_tool
from gaphas.tool.rubberband import rubberband_tool
from gaphas.tool.scroll import scroll_tool, scroll_tools
from gaphas.tool.viewfocus import view_focus_tool
from gaphas.tool.zoom import zoom_tool, zoom_tools
